#!/usr/bin/env python

import pdb
import numpy as np
import os

from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from common import get_dict
from get_database import Database, input_args
from vtk_functions import read_geo, write_geo


def main(db, geometries):
    steps_new = {'0074_0001': 600,
                 '0075_1001': 1600,
                 '0090_0001': 1200,
                 '0063_0001': 5750,
                 '0064_0001': 5750,
                 '0065_0001': 5750,
                 '0075_0001': 5750,
                 '0076_0001': 5750,
                 '0068_0001': 2000,
                 '0099_0001': 1000,
                 '0092_0001':  850}

    # get bcs of all models
    bc_all = get_dict(db.db_params)

    for geo in steps_new.keys():
        # 3d exported time
        time, _ = db.get_inflow(geo)
        if time is None:
            continue

        # number of steps
        numstep = db.get_3d_numstep(geo)

        # output increment
        nt_out = db.get_3d_increment(geo)

        if not numstep/nt_out % 1 == 0.0:
            print(geo + ' output step not integer: ' + str(numstep/nt_out), nt_out, numstep, steps_new[geo])
            bc_all[geo]['params']['sim_steps_per_cycle'] = str(steps_new[geo])
    np.save(db.db_params, bc_all)


if __name__ == '__main__':
    descr = 'Fix wrong time step / numstep'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import numpy as np
import os

from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from get_database import Database, input_args
from vtk_functions import read_geo, write_geo


def main(db, geometries):
    for geo in geometries:
        print('Fixing geometry ' + geo)
        folder_out = os.path.join(db.fpath_gen, 'surfaces', geo)

        f_all = os.path.join(folder_out, 'all_exterior.vtp')
        f_wall = os.path.join(folder_out, 'wall.vtp')

        read_all = read_geo(f_all).GetOutput()
        read_wall = read_geo(f_wall).GetOutput()

        # read indices
        fid_all = v2n(read_all.GetCellData().GetArray('BC_FaceID'))
        eid_all = v2n(read_all.GetCellData().GetArray('GlobalElementID'))
        eid_wall = v2n(read_wall.GetCellData().GetArray('GlobalElementID'))

        # find wall indices in all_exterior
        index = np.argsort(eid_all)
        search = np.searchsorted(eid_all[index], eid_wall)
        ind_wall = index[search]

        wall_id = np.unique(fid_all[ind_wall])
        assert wall_id.shape[0] == 1, 'wall BC_FaceID not unique'

        if wall_id[0] == 0:
            print('  Nothing to do')
            continue

        # change id
        fid_all[ind_wall] = 0

        # export
        out_array = n2v(fid_all)
        out_array.SetName('BC_FaceID')
        read_all.GetCellData().RemoveArray('BC_FaceID')
        read_all.GetCellData().AddArray(out_array)
        write_geo(f_all, read_all)


if __name__ == '__main__':
    descr = 'Fix wrong surface id'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import pdb
import numpy as np

from vtk_functions import read_geo, write_geo

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n


def main():
    f_in = '/home/pfaller/work/osmsc/data_generated/centerlines/0075_1001.vtp'
    f_out = '0075_1001.vtp'

    # read geometry from file
    cent = read_geo(f_in).GetOutput()

    # get number of points
    n_point = cent.GetNumberOfPoints()

    # create dummy array
    array = 42 * np.ones(n_point)

    # add array to centerline
    out_array = n2v(array)
    out_array.SetName('Answer to the Ultimate Question of Life, The Universe, and Everything')
    cent.GetPointData().AddArray(out_array)

    # write geometry to file
    write_geo(f_out, cent)

    # set break-point for debugging
    pdb.set_trace()


if __name__ == '__main__':
    main()



import os
import vtk
import numpy as np
from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from vtk_functions import read_geo, write_geo, add_array


def get_velocities(geo):
    narrs = geo.GetPointData().GetNumberOfArrays()
    velocities = []
    count = 0
    for ind in range(narrs):
        name = geo.GetPointData().GetArrayName(ind)
        if 'velocity_' in name and not 'average_' in name:
            if count % 1 == 0:
                print(name)
                cur_velocity = v2n(geo.GetPointData().GetArray(ind))
                velocities.append(np.reshape(cur_velocity, cur_velocity.size))
                print(count)
            count = count + 1
    velocities = np.array(velocities)
    return velocities
    

if __name__ == "__main__":
    geo = read_geo('0069_0001.vtu')
    velocities = get_velocities(geo).transpose()
    svd = np.linalg.svd(velocities, full_matrices=False)
    lerror = 0
    sol_index = 0
    # find maximum error
    for i in range(velocities.shape[1]):
        A = np.matmul(np.diag(svd[1]),svd[2][:,i])
        C = svd[0][:,0] * A[0]
        error = np.linalg.norm(velocities[:,i] - C)
        if error > lerror:
            error = lerror
            sol_index = i
    print(error)
    print(i)
    ref_solution = np.copy(velocities[:,sol_index])
    A = np.matmul(np.diag(svd[1]),svd[2][:,sol_index])
    print(A[0:10])
    C1 = svd[0][:,0] * A[0]
    C2 = np.matmul(svd[0][:,0:2], A[0:2])
    C5 = np.matmul(svd[0][:,0:5], A[0:5])
    C10 = np.matmul(svd[0][:,0:10], A[0:10])
    node_all = geo.GetPointData()
    for i in range(node_all.GetNumberOfArrays() - 1):
        node_all.RemoveArray(1)
    ncomps = int(svd[0].shape[0]/3)
    add_array(geo, 'approx_1 modes', np.reshape(C1,(ncomps,3)))
    add_array(geo, 'approx_2 modes', np.reshape(C2,(ncomps,3)))
    add_array(geo, 'approx_5 modes', np.reshape(C5,(ncomps,3)))
    add_array(geo, 'approx_10 modes', np.reshape(C10,(ncomps,3)))
    add_array(geo, 'solution', np.reshape(ref_solution,(ncomps,3)))
    write_geo('test.vtu', geo)


#!/usr/bin/env python

import os
import re
import vtk
import argparse
import pdb

import numpy as np
import matplotlib.pyplot as plt

from collections import defaultdict

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

from get_database import Database
from vtk_functions import read_geo, write_geo, get_all_arrays
from simulation_io import get_dict


fsize = 20
plt.rcParams.update({'font.size': fsize})

plt.rcParams.update({'text.usetex': True, 'font.family': 'serif', 'font.serif': 'Computer Modern Roman'})

def plot_area():
    db = Database()
    f_in = db.get_centerline_path('0003_0001')
    f_out = os.path.join('.', 'OSMSC_0003_0001_branch0')

    # get centerline
    geo = read_geo(f_in).GetOutput()
    arrays, _ = get_all_arrays(geo)

    # extract branch
    br = 0
    mask = arrays['BranchId'] == br

    # get plot quantities
    path = arrays['Path'][mask]
    area_slice = arrays['CenterlineSectionArea'][mask]
    area_vmtk = arrays['MaximumInscribedSphereRadius'][mask] ** 2 * np.pi

    print('factor', area_slice[0] / area_vmtk[0])

    # make plot
    fig, ax = plt.subplots(dpi=300, figsize=(6, 6))
    ax.plot(path, area_slice, 'r-')
    ax.plot(path, area_vmtk, 'b-')
    ax.legend(['Area from slicing', 'Area from MISR'])
    ax.set_xlim(left=0)
    ax.set_xticks([0, 2, 4])
    ax.set_xticklabels(['Inlet', '2', '4'])
    plt.xlabel('Branch path [cm]')
    plt.ylabel('Area [cm$^2$]')
    plt.grid()
    fig.savefig(f_out, bbox_inches='tight')


def plot_model_statistics(db, geometries_paper):

    pie = defaultdict(lambda: defaultdict(int))
    cats = ['deliverable_category', 'vascular_state', 'treatment', 'image_data_modality', 'paper_reference', 'gender']

    names = {'deliverable_category': 'Vascular anatomy',
             'vascular_state': 'Vascular state',
             'treatment': 'Treatment',
             'image_data_modality': 'Imaging',
             'paper_reference': 'Literature reference',
             'gender': 'Gender'}

    # count all categories
    for geo in geometries_paper:
        bc_def = db.get_bcs(geo)
        if bc_def is not None:
            pie['has_bc']['yes'] += 1
            params = bc_def['params']
            for cat in cats:
                if cat in params and 'unpublished' not in params[cat]:
                    if cat == 'paper_reference':
                        name = params[cat][:-2]
                    else:
                        name = params[cat].capitalize()
                elif cat in params and params[cat] == 'Unclassified':
                    name = 'Normal'
                else:
                    if cat == 'paper_reference':
                        name = 'Unpublished'
                    else:
                        name = 'None'
                pie[cat][name] += 1
        else:
            pie['has_bc']['no'] += 1

    # make plots
    fig, axs = plt.subplots(2, 2, dpi=300, figsize=(25, 15))

    selection = ['deliverable_category', 'vascular_state', 'treatment', 'paper_reference']
    for cat, ax in zip(selection, axs.ravel()):
        labels = np.array([re.sub(r'\([^)]*\)', '', c) for c in pie[cat].keys()])
        sizes = np.array(list(pie[cat].values()))
        order = np.argsort(sizes)

        print('num', np.sum(sizes))
        abs_size = lambda p: '{:.0f}'.format(p * np.sum(sizes) / 100)

        theme = plt.get_cmap('Reds')
        ax.set_prop_cycle("color", [theme(1. * i / len(sizes)) for i in range(len(sizes))])
        ax.pie(sizes[order], labels=labels[order], autopct=abs_size)
        ax.axis('equal')
        ax.set_title(names[cat], fontsize=40, pad=20)

    f_out = os.path.join(db.get_statistics_dir(), 'repo_statistics')#.pgf
    fig.savefig(f_out, bbox_inches='tight')
    plt.close(fig)


def plot_collage(db, geos):
    nx = 10
    ny = 8
    assert nx * ny >= len(geos), 'choose larger image grid: ' + str(len(geos))
    fig , ax = plt.subplots(nx, ny, figsize=(ny * 2, nx * 2.5), dpi=100)
    ig = 0
    for i in range(nx):
        for j in range(ny):
            ax[i, j].axis('off')
            if ig >= len(geos):
                continue
            geo = geos[ig]
            impath = db.get_png(geo)
            if not os.path.exists(impath):
                continue
            im = plt.imread(impath)
            ax[i, j].imshow(im)
            ax[i, j].set_title(geo.replace('_', '\_'), fontsize=18)
            ig += 1
    f_out = os.path.join(db.get_statistics_dir(), 'repo_models')#.pgf
    #fig.tight_layout(pad=3.0)
    fig.savefig(f_out, bbox_inches='tight')
    plt.close(fig)

def main():
    db = Database('1spb_length_stenosis')
    # geometries_paper = db.get_geometries_select('paper')

    # only geometries where a 0d AND a 3d_rerun solution exisis
    geometries_paper = []
    for geo in sorted(list(get_dict(db.get_log_file_0d()).keys())):
        if os.path.exists('/home/pfaller/work/osmsc/studies/ini_1d_quad/3d_flow/' + geo + '.vtp'):
            geometries_paper += [geo]
        else:
            if os.path.exists('/home/pfaller/work/osmsc/studies/ini_zero/3d_flow/' + geo + '.vtp'):
                geometries_paper += [geo]

    print(geometries_paper)
    plot_collage(db, geometries_paper)
    plot_model_statistics(db, geometries_paper)
    plot_area()


if __name__ == '__main__':
    descr = 'Make plots for 3D-1D-0D paper'
    main()



#!/usr/bin/env python

import pdb
import os
import shutil
import glob
import matplotlib.cm as cm
from collections import OrderedDict, defaultdict

import numpy as np
import scipy.interpolate

from get_database import input_args, Database, SVProject, SimVascular, Post
from common import coronary_sv_to_oned, get_dict
from get_sv_meshes import get_meshes


class Project:
    def __init__(self, db, geo, mode):
        self.db = db
        self.geo = geo
        self.mode = mode
        self.opt = self.get_sv_opt()

    def create_sv_project(self):
        print('Estimated cycles: ' + str(self.opt['n_cycle']))
        self.check_files()
        self.make_folders()
        self.write_inflow('1d')
        self.write_inflow('3d')
        self.copy_files()
        self.write_svproj_file()
        self.write_model()
        self.write_simulation()
        self.write_mesh()
        self.write_pre()
        self.write_solver()
        self.write_bc(os.path.join(self.db.get_solve_dir_3d(self.geo)), False)
        self.write_path_segmentation()

    def get_sv_opt(self):
        # get boundary conditions
        bc_def = self.db.get_bcs(self.geo)
        if 'steady' in self.mode:
            bc_def = rc_to_r(bc_def)

        # number of cycles
        n_cycle = self.estimate_n_cycle()

        # number of time steps
        numstep = self.db.get_3d_numstep(self.geo)
    
        # time step
        dt = self.db.get_3d_timestep(self.geo)
    
        # time increment
        nt_out = self.db.get_3d_increment(self.geo)
    
        # read inflow conditions
        time, inflow = self.db.get_inflow_smooth(self.geo)

        if self.mode == '':
            pass
        elif self.mode == 'ini_1d_quad':
            n_cycle = 5
        elif self.mode == 'steady':
            n_cycle = 1
            numstep = 100
            nt_out = 1
            inflow = np.mean(inflow) * np.ones(len(time))
        elif self.mode == 'steady0':
            n_cycle = 1
            inflow = inflow[0] * np.ones(len(time))
        elif self.mode == 'irene':
            n_cycle = 1
    
            # sample inflow at fine time step
            dt_fine = dt# * 1e-3
            time_fine = np.arange(0, time[-1] + dt_fine, dt_fine)
            inflow_fine = scipy.interpolate.interp1d(time, inflow, fill_value="extrapolate")(time_fine)
    
            # find (first) time where inflow reaches mean flow
            f_mean = np.mean(inflow)
            t_mean = np.argwhere(np.diff(np.sign(inflow_fine - f_mean))).flatten()[0]
            assert np.abs(inflow_fine[t_mean]/f_mean - 1) < 5e-2, 'could not find mean-flow'
    
            # shift inflow to start at mean flow
            inflow_fine_shift = np.roll(inflow_fine, -t_mean)
    
            # interpolate back to original time step
            inflow = scipy.interpolate.interp1d(time_fine, inflow_fine_shift)(time)
            assert np.abs(inflow[0]/f_mean - 1) < 5e-2, 'inflow does not start at mean flow'
    
            p_str = 'time post nt=' + str(numstep-t_mean)
            p_str += ' mean inflow f=' + str(f_mean * Post().convert['flow']) + ' [l/]'
            p_str += ' at t=' + str(t_mean * dt_fine) + ' s'
            print(p_str)
        else:
            raise ValueError('Unknown mode ' + self.mode)

        # number of time steps
        n_time = n_cycle * numstep + nt_out
    
        # create inflow string
        inflow_str = array_to_sv(np.vstack((time, inflow)).T)
    
        # number of fourier modes
        n_t = time.shape[0]
        if n_t % 2 == 0:
            n_fourier = int(n_t / 2 + 1)
        else:
            n_fourier = int((n_t + 1) / 2)
        n_fourier = np.min([200, n_fourier])

        # number of points
        n_point = len(time)
    
        # get inflow type
        inflow_type = bc_def['bc']['inflow']['type']
        if inflow_type == 'womersley':
            inflow_type = 'plug'
        if inflow_type not in ['parabolic', 'plug', 'womersley']:
            raise ValueError('unknown inflow type ' + inflow_type)

        # set svsolver options
        opt = {'bc': bc_def,
               'density': '1.06',
               'viscosity': '0.04',
               'backflow': '0.2',
               'advection': 'Convective',
               'inflow': 'inflow.flow',
               'inflow_data': np.vstack((time, inflow)).T,
               'inflow_str': inflow_str,
               'inflow_type': inflow_type,
               'fourier_modes': str(n_fourier),
               'fourier_period': str(time[-1]),
               'fourier_points': str(n_point),
               'max_iter_continuity': '400',
               'max_iter_momentum': '10',
               'max_iter_ns_solver': '10',
               'min_iter': '3',
               'num_krylov': '300',
               'num_solve': '1',
               'num_time': str(int(n_time)),
               'num_restart': '1',#str(nt_out),
               'n_cycle': n_cycle,
               'bool_surf_stress': 'True',
               'coupling': 'Implicit',
               'print_avg_sol': 'True',
               'print_err': 'False',
               'quad_boundary': '3',
               'quad_interior': '2',
               'residual_control': 'True',
               'residual_criteria': '0.01',
               'residual_tolerance': '1.0e12',
               'step_construction': '5',
               'time_int_rho': '0.5',
               'time_int_rule': 'Second Order',
               'time_step': str(dt),
               'tol_continuity': '0.01',
               'tol_momentum': '0.01',
               'tol_ns_solver': '0.01',
               'svls_type': 'NS',
               'mesh_initial': os.path.join('mesh-complete', 'initial.vtu'),
               'mesh_vtu': os.path.join('mesh-complete', 'mesh-complete.mesh.vtu'),
               'mesh_vtp': os.path.join('mesh-complete', 'mesh-complete.exterior.vtp'),
               'mesh_inflow': os.path.join('mesh-complete', 'mesh-surfaces', 'inflow.vtp'),
               'mesh_walls': os.path.join('mesh-complete', 'walls_combined.vtp')}
        return opt
    
    def estimate_n_cycle(self, n_extra=1, n_default=5):
        # get numerical time constants
        db = Database('1spb_length')
        res_num = get_dict(db.get_convergence_path())

        params = db.get_bcs(self.geo)
        cat = params['params']['deliverable_category']

        # pure resistance
        bc_types = np.unique(list(params['bc_type'].values()))
        if len(bc_types) == 1 and bc_types[0] == 'resistance':
            return 2

        # time constant
        if self.geo not in res_num:
            tau_list = []
            for outlet, val in params['bc'].items():
                if outlet in params['bc_type'] and params['bc_type'][outlet] == 'rcr':
                    tau_list += [val['C'] * val['Rd']]
            tau = np.max(tau_list)
        else:
            tau = np.mean(res_num[self.geo]['tau']['pressure'])

        # tolerance for asymptotic convergence
        tol = 0.01

        # number of cardiac cycles required to reach tolerance (+ extra) from zero ICs
        n_cycle_zero = int(- np.log(tol) * tau + 0.5) + n_extra

        # estimated number of cycles from steady state ICs
        n_cycle_mean = int(4 + 2 * (tau + 0.5))

        if cat == 'Coronary':
            return n_cycle_zero
        else:
            return np.min([n_cycle_zero, n_cycle_mean])
    
    def write_svproj_file(self):
        t = str(self.db.svproj.t)
        proj_head = ['<?xml version="1.0" encoding="UTF-8"?>',
                     '<projectDescription version="1.0">']
        proj_end = ['</projectDescription>']
    
        with open(self.db.get_svproj_file(self.geo), 'w+') as f:
            # write header
            for s in proj_head:
                f.write(s + '\n')
    
            # write images/segmentations
            for k, s in self.db.svproj.dir.items():
                f.write(t + '<' + k + ' folder_name="' + s + '"')
                if k == 'images' and self.db.get_img(self.geo) is not None:
                    img = os.path.basename(self.db.get_img(self.geo))
                    f.write('>\n')
                    f.write(
                        t * 2 + '<image name="' + os.path.splitext(img)[0] + '" in_project="yes" path="' + img + '"/>\n')
                    f.write(t + '</' + k + '>\n')
                # elif k == 'segmentations':
                else:
                    f.write('/>\n')
    
            # write end
            for s in proj_end:
                f.write(s + '\n')
    
    def write_model(self):
        t = str(self.db.svproj.t)
        model_head = ['<?xml version="1.0" encoding="UTF-8" ?>',
                      '<format version="1.0" />',
                      '<model type="PolyData">',
                      t + '<timestep id="0">',
                      t * 2 + '<model_element type="PolyData" num_sampling="0">']
        model_end = [t * 3 + '<blend_radii />',
                     t * 3 + '<blend_param blend_iters="2" sub_blend_iters="3" cstr_smooth_iters="2" lap_smooth_iters="50" '
                             'subdivision_iters="1" decimation="0.01" />',
                     t * 2 + '</model_element>',
                     t + '</timestep>',
                     '</model>']
    
        # read boundary conditions
        bc_def = self.opt['bc']
        bc_def['preid']['wall'] = 0
    
        # get cap names
        caps = self.db.get_surface_names(self.geo, 'caps')
        caps += ['wall']
    
        # sort caps according to face id
        ids = np.array([repr(int(float(bc_def['preid'][c] + 1))) for c in caps])
        order = np.argsort(ids)
        caps = np.array(caps)[order]
        ids = ids[order]
    
        # display colors for caps
        colors = cm.jet(np.linspace(0, 1, len(caps)))
    
        # write model file
        with open(self.db.get_svproj_mdl_file(self.geo), 'w+') as f:
            # write header
            for s in model_head:
                f.write(s + '\n')
    
            #             <segmentations>
            #                 <seg name="aorta_final_new" />
            #                 <seg name="btrunk_final" />
            #                 <seg name="carotid_final" />
            #                 <seg name="subclavian_final_new" />
            #             </segmentations>
    
            # write faces
            f.write(t * 3 + '<faces>\n')
            for i, c in enumerate(caps):
                c_str = t * 4 + '<face id="' + ids[i] + '" name="' + c + '" type='
                if c == 'wall':
                    c_str += '"wall"'
                else:
                    c_str += '"cap"'
                for j in range(3):
                    c_str += ' color' + repr(j + 1) + '="' + repr(colors[i, j]) + '"'
                f.write(c_str + ' visible="true" opacity="1" />\n')
            f.write(t * 3 + '</faces>\n')
    
            # write end
            for s in model_end:
                f.write(s + '\n')
    
    def write_path_segmentation(self):
        # SimVascular instance
        sv = SimVascular()
    
        # get paths
        p = OrderedDict()
        p['f_path_in'] = self.db.get_path_file(self.geo)
        p['f_path_out'] = os.path.join(self.db.get_svproj_dir(self.geo), self.db.svproj.dir['paths'])
    
        seg_dir = self.db.get_seg_dir(self.geo)
        segments = glob.glob(os.path.join(seg_dir, '*'))
    
        err_seg = ''
        for s in segments:
            p['f_seg_in'] = s
            p['f_seg_out'] = os.path.join(self.db.get_svproj_dir(self.geo), self.db.svproj.dir['segmentations'])
    
            if '.tcl' in s:
                continue
    
            # assemble call string
            sv_string = [os.path.join(os.getcwd(), 'sv_get_path_segmentation.py')]
            for v in p.values():
                sv_string += [v]
    
            err = sv.run_python_legacyio(sv_string)[1]
            if err:
                err_seg += os.path.basename(s).split('.')[0] + '\n'
    
        # execute SimVascular-Python
        return err_seg
    
    def write_mesh(self):
        t = str(self.db.svproj.t)
        mesh_generic = ['<?xml version="1.0" encoding="UTF-8" ?>',
                        '<format version="1.0" />',
                        '<mitk_mesh type="TetGen" model_name="' + self.geo + '">',
                        t + '<timestep id="0">',
                        t * 2 + '<mesh type="TetGen">',
                        t * 3 + '<command_history>',
                        t * 4 + '<command content="option surface 1" />',
                        t * 4 + '<command content="option volume 1" />',
                        t * 4 + '<command content="option UseMMG 1" />',
                        t * 4 + '<command content="setWalls" />',
                        # t * 4 + '<command content="option Optimization 3" />',
                        # t * 4 + '<command content="option QualityRatio 1.4" />',
                        t * 4 + '<command content="option NoBisect" />',
                        # t * 4 + '<command content="AllowMultipleRegions 0" />',
                        t * 4 + '<command content="generateMesh" />',
                        t * 4 + '<command content="writeMesh" />',
                        t * 3 + '</command_history>',
                        t * 2 + '</mesh>',
                        t + '</timestep>',
                        '</mitk_mesh>']
    
        fname = os.path.join(self.db.get_svproj_dir(self.geo), self.db.svproj.dir['meshes'], self.geo + '.msh')
    
        # write generic mesh file
        with open(fname, 'w+') as f:
            for s in mesh_generic:
                f.write(s + '\n')
    
    def write_inflow(self, model, n_mode=10, n_sample_real=256):
        # read inflow conditions
        opt = self.opt
        time = opt['inflow_data'][:, 0]
        inflow = opt['inflow_data'][:, 1]

        if time is None:
            raise ValueError('no inflow')

        # save inflow file
        fpath = self.db.get_sv_flow_path(self.geo, model)
        os.makedirs(os.path.dirname(fpath), exist_ok=True)

        # reverse flow for svOneDSolver
        fmt = '%.10e'
        if model == '1d':
            np.savetxt(fpath, np.vstack((time, - inflow)).T, fmt=fmt)
        else:
            np.savetxt(fpath, np.vstack((time, inflow)).T, fmt=fmt)

        return len(inflow), time[-1]
    
    def write_pre(self, solver='svsolver'):
        """
        Create input file for svpre
        """
        # get boundary conditions
        bc_def = self.opt['bc']
    
        # read inflow conditions
        time, _ = self.db.get_inflow_smooth(self.geo)
    
        # outlet names
        outlets = self.db.get_outlet_names(self.geo)
    
        # get solver options
        opt = self.opt
    
        with open(self.db.get_svpre_file(self.geo, solver), 'w+') as f:
            # enter debug mode
            # f.write('verbose true\n')
    
            # write volume mesh
            f.write('mesh_and_adjncy_vtu ' + opt['mesh_vtu'] + '\n')
    
            # write surface mesh
            fpath_surf = os.path.join('mesh-complete', 'mesh-surfaces')
    
            # write surfaces (sort according to surface ID for readability)
            f.write('set_surface_id_vtp ' + opt['mesh_vtp'] + ' 1\n')
            f.write('set_surface_id_vtp ' + opt['mesh_inflow'] + ' 2\n')
            for k in outlets:
                v = bc_def['preid'][k] + 1
                if int(v) > 1:
                    f_surf = os.path.join(fpath_surf, k + '.vtp')
    
                    # check if mesh file exists
                    f_surf_full = os.path.join(self.db.get_solve_dir_3d(self.geo), f_surf)
                    assert os.path.exists(f_surf_full), 'file ' + f_surf + ' does not exist'
    
                    f.write('set_surface_id_vtp ' + f_surf + ' ' + repr(int(v)) + '\n')
            f.write('\n')
    
            if solver == 'perigee':
                return
    
            # write inlet bc
            f.write('prescribed_velocities_vtp ' + opt['mesh_inflow'] + '\n\n')
    
            # generate inflow
            f.write('bct_analytical_shape ' + opt['inflow_type'] + '\n')
            f.write('bct_period ' + opt['fourier_period'] + '\n')
            f.write('bct_point_number ' + opt['fourier_points'] + '\n')
            f.write('bct_fourier_mode_number ' + opt['fourier_modes'] + '\n')
            # f.write('bct_create ' + opt['mesh_inflow'] + ' ' + self.db.get_sv_flow_path_rel(self.geo, '3d_constant') + '\n')
            f.write('bct_create ' + opt['mesh_inflow'] + ' ' + opt['inflow'] + '\n')
            f.write('bct_write_dat bct.dat\n')
            f.write('bct_write_vtp bct.vtp\n\n')
    
            # write default parameters
            f.write('fluid_density ' + opt['density'] + '\n')
            f.write('fluid_viscosity ' + opt['viscosity'] + '\n\n')
    
            # reference pressure
            for cap in outlets:
                bc = bc_def['bc'][cap]
                if cap == 'inflow' or cap == 'wall':
                    continue
                if 'Po' in bc and bc_def['bc_type'][cap] == 'resistance':
                    p0 = str(bc['Po'])
                else:
                    p0 = '0.0'
                f.write('pressure_vtp ' + os.path.join(fpath_surf, cap + '.vtp') + ' ' + p0 + '\n')
            f.write('\n')
    
            # set previous results as initial condition
            f.write('read_pressure_velocity_vtu ' + opt['mesh_initial'] + '\n\n')
            # f.write('initial_pressure 0\n')
            # f.write('initial_velocity 0.0001 0.0001 0.0001\n\n')
    
            # no slip boundary condition
            f.write('noslip_vtp ' + opt['mesh_walls'] + '\n\n')
    
            # request outputs
            f.write('write_geombc geombc.dat.1\n')
            f.write('write_restart restart.0.1\n')
            f.write('write_numstart 0\n\n')
    
        # write start file
        fname_start = os.path.join(self.db.get_solve_dir_3d(self.geo), 'numstart.dat')
        with open(fname_start, 'w+') as f:
            f.write('0')
    
    def write_solver(self):
        # get boundary conditions
        bc_def = self.opt['bc']
    
        # ordered outlets
        outlets = self.db.get_outlet_names(self.geo)
    
        # get solver options
        opt = self.opt
    
        with open(self.db.get_solver_file(self.geo), 'w+') as f:
            # write default parameters
            # todo: get from tcl
            f.write('Density: ' + opt['density'] + '\n')
            f.write('Viscosity: ' + opt['viscosity'] + '\n\n')
    
            # time step
            f.write('Number of Timesteps: ' + opt['num_time'] + '\n')
            f.write('Time Step Size: ' + opt['time_step'] + '\n\n')
    
            # output
            f.write('Number of Timesteps between Restarts: ' + opt['num_restart'] + '\n')
            f.write('Number of Force Surfaces: 1\n')
            f.write('Surface ID\'s for Force Calculation: 0\n')
            f.write('Force Calculation Method: Velocity Based\n')
            f.write('Print Average Solution: ' + opt['print_avg_sol'] + '\n')
            f.write('Print Error Indicators: ' + opt['print_err'] + '\n\n')
    
            f.write('Time Varying Boundary Conditions From File: True\n\n')
    
            f.write('Step Construction:')
            for i in range(int(opt['step_construction'])):
                f.write(' 0 1')
            f.write('\n\n')
    
            # collect faces for each boundary condition type
            bc_ids = defaultdict(list)
            for cap in outlets:
                bc_ids[bc_def['bc_type'][cap]] += [int(bc_def['preid'][cap]) + 1]
    
            # boundary conditions
            names = {'rcr': 'RCR', 'resistance': 'Resistance', 'coronary': 'COR'}
            for t, v in bc_ids.items():
                f.write('Number of ' + names[t] + ' Surfaces: ' + str(len(v)) + '\n')
                f.write('List of ' + names[t] + ' Surfaces: ' + str(v).replace(',', '')[1:-1] + '\n')
    
                if t == 'rcr' or t == 'coronary':
                    f.write(names[t] + ' Values From File: True\n\n')
                elif t == 'resistance':
                    f.write('Resistance Values: ')
                    for cap in self.db.get_outlet_names(self.geo):
                        if bc_def['bc_type'][cap] == 'resistance':
                            f.write(str(bc_def['bc'][cap]['R']) + ' ')
                    f.write('\n\n')
                else:
                    raise ValueError('Boundary condition ' + t + ' unknown')
    
            f.write('Pressure Coupling: ' + opt['coupling'] + '\n')
            f.write('Number of Coupled Surfaces: ' + str(len(bc_def['bc']) - 2) + '\n\n')
    
            f.write('Backflow Stabilization Coefficient: ' + opt['backflow'] + '\n')
    
            # nonlinear solver
            f.write('Residual Control: ' + opt['residual_control'] + '\n')
            f.write('Residual Criteria: ' + opt['residual_criteria'] + '\n')
            f.write('Residual Tolerance: ' + opt['residual_tolerance'] + '\n')
            f.write('Minimum Required Iterations: ' + opt['min_iter'] + '\n')
    
            # linear solver
            f.write('svLS Type: ' + opt['svls_type'] + '\n')
            f.write('Number of Krylov Vectors per GMRES Sweep: ' + opt['num_krylov'] + '\n')
            f.write('Number of Solves per Left-hand-side Formation: ' + opt['num_solve'] + '\n')
    
            f.write('Tolerance on Momentum Equations: ' + opt['tol_momentum'] + '\n')
            f.write('Tolerance on Continuity Equations: ' + opt['tol_continuity'] + '\n')
            f.write('Tolerance on svLS NS Solver: ' + opt['tol_ns_solver'] + '\n')
    
            f.write('Maximum Number of Iterations for svLS NS Solver: ' + opt['max_iter_ns_solver'] + '\n')
            f.write('Maximum Number of Iterations for svLS Momentum Loop: ' + opt['max_iter_momentum'] + '\n')
            f.write('Maximum Number of Iterations for svLS Continuity Loop: ' + opt['max_iter_continuity'] + '\n')
    
            # time integration
            f.write('Time Integration Rule: ' + opt['time_int_rule'] + '\n')
            f.write('Time Integration Rho Infinity: ' + opt['time_int_rho'] + '\n')
    
            f.write('Flow Advection Form: ' + opt['advection'] + '\n')
    
            f.write('Quadrature Rule on Interior: ' + opt['quad_interior'] + '\n')
            f.write('Quadrature Rule on Boundary: ' + opt['quad_boundary'] + '\n')
    
    def write_simulation(self):
        # get boundary conditions
        bc_def = self.opt['bc']

        # get outlet names
        outlets = self.db.get_outlet_names(self.geo)
    
        # get solver options
        opt = self.opt
    
        # tab
        t = str(self.db.svproj.t)
    
        sim_header = ['<?xml version="1.0" encoding="UTF-8" ?>',
                      '<format version="1.0" />',
                      '<mitk_job model_name="' + self.geo + '" mesh_name="' + self.geo + '" status="Simulation failed">',
                      t + '<job>']
    
        basic_props = [['Fluid Density', opt['density']],
                       ['Fluid Viscosity', opt['viscosity']],
                       ['IC File', opt['mesh_initial']],
                       ['Initial Pressure', '0'],
                       ['Initial Velocities', '0.0001 0.0001 0.0001']]
    
        inflow_props = [['Analytic Shape', bc_def['bc']['inflow']['type']],
                        ['BC Type', 'Prescribed Velocities'],
                        ['Flip Normal', 'False'],
                        ['Flow Rate', opt['inflow_str']],
                        ['Fourier Modes', opt['fourier_modes']],
                        ['Original File', 'inflow.flow'],
                        ['Period', opt['fourier_period']],
                        ['Point Number', opt['fourier_points']]]
    
        wall_props = [['Type', 'rigid']]
    
        solver_props = [['Backflow Stabilization Coefficient', opt['backflow']],
                        ['Flow Advection Form', opt['advection']],
                        ['Force Calculation Method', 'Velocity Based'],
                        ['Maximum Number of Iterations for svLS Continuity Loop', opt['max_iter_continuity']],
                        ['Maximum Number of Iterations for svLS Momentum Loop', opt['max_iter_momentum']],
                        ['Maximum Number of Iterations for svLS NS Solver', opt['max_iter_ns_solver']],
                        ['Minimum Required Iterations', opt['min_iter']],
                        ['Number of Krylov Vectors per GMRES Sweep', opt['num_krylov']],
                        ['Number of Solves per Left-hand-side Formation', opt['num_solve']],
                        ['Number of Timesteps', opt['num_time']],
                        ['Number of Timesteps between Restarts', opt['num_restart']],
                        ['Output Surface Stress', opt['bool_surf_stress']],
                        ['Pressure Coupling', opt['coupling']],
                        ['Print Average Solution', opt['print_avg_sol']],
                        ['Print Error Indicators', opt['print_err']],
                        ['Quadrature Rule on Boundary', opt['quad_boundary']],
                        ['Quadrature Rule on Interior', opt['quad_interior']],
                        ['Residual Control', opt['residual_control']],
                        ['Residual Criteria', opt['residual_criteria']],
                        ['Residual Tolerance', opt['residual_tolerance']],
                        ['Step Construction', opt['step_construction']],
                        ['Time Integration Rho Infinity', opt['time_int_rho']],
                        ['Time Integration Rule', opt['time_int_rule']],
                        ['Time Step Size', opt['time_step']],
                        ['Tolerance on Continuity Equations', opt['tol_continuity']],
                        ['Tolerance on Momentum Equations', opt['tol_momentum']],
                        ['Tolerance on svLS NS Solver', opt['tol_ns_solver']],
                        ['svLS Type', opt['svls_type']]]
    
        run_props = [['Number of Processes', '8']]
    
        with open(self.db.get_svproj_sjb_file(self.geo), 'w+') as f:
            for h in sim_header:
                f.write(h + '\n')
    
            f.write(t * 2 + '<basic_props>\n')
            print_props(f, basic_props, t * 3)
            f.write(t * 2 + '</basic_props>\n')
    
            # bcs
            f.write(t * 2 + '<cap_props>\n')
    
            # outflow
            for k in outlets:
                f.write(t * 3 + '<cap name="' + k + '">\n')
    
                tp = bc_def['bc_type'][k]
                bc = bc_def['bc'][k]
    
                if tp == 'rcr':
                    rcr_val = ' '.join([str(bc[v]) for v in ['Rp', 'C', 'Rd']])
                    f.write(t * 4 + '<prop key="BC Type" value="RCR" />\n')
                    f.write(t * 4 + '<prop key="C Values" value="" />\n')
                    if 'Po' in bc:
                        f.write(t * 4 + '<prop key="Pressure" value="' + str(bc['Po']) + '" />\n')
                    else:
                        f.write(t * 4 + '<prop key="Pressure" value="0.0" />\n')
                    f.write(t * 4 + '<prop key="R Values" value="" />\n')
                    f.write(t * 4 + '<prop key="Values" value="' + rcr_val + '" />\n')
                elif tp == 'resistance':
                    f.write(t * 4 + '<prop key="BC Type" value="Resistance" />\n')
                    if 'Po' in bc:
                        f.write(t * 4 + '<prop key="Pressure" value="' + str(bc['Po']) + '" />\n')
                    else:
                        f.write(t * 4 + '<prop key="Pressure" value="0.0" />\n')
                    f.write(t * 4 + '<prop key="Values" value="' + str(bc['R']) + '" />\n')
                elif tp == 'coronary':
                    # convert parameters to SimVascular format
                    bc_sv = coronary_sv_to_oned(bc)
    
                    c_val = ' '.join([str(bc_sv[v]) for v in ['Ca', 'Cc']])
                    r_val = ' '.join([str(bc_sv[v]) for v in ['Ra1', 'Ra2', 'Rv1']])
                    p_val = str(bc_sv['P_v'])
                    a_val = ' '.join([str(bc_sv[v]) for v in ['Ra1', 'Ca', 'Ra2', 'Cc', 'Rv1']])
    
                    p_v = bc_def['coronary'][bc['Pim']]
    
                    # save ventricular pressure to file
                    f_out = os.path.join(self.db.get_solve_dir_3d(self.geo), bc['Pim'])
                    np.savetxt(f_out, p_v)
    
                    f.write(t * 4 + '<prop key="BC Type" value="Coronary" />\n')
                    f.write(t * 4 + '<prop key="C Values" value="' + c_val + '" />\n')
                    f.write(t * 4 + '<prop key="Original File" value="' + os.path.join(self.geo, bc['Pim']) + '" />\n')
                    f.write(t * 4 + '<prop key="Pressure" value="' + p_val + '" />\n')
                    f.write(t * 4 + '<prop key="Pressure Period" value="' + str(p_v[-1, 0]) + '" />\n')
                    f.write(t * 4 + '<prop key="Pressure Scaling" value="1.0" />\n')
                    f.write(t * 4 + '<prop key="R Values" value="' + r_val + '" />\n')
                    f.write(t * 4 + '<prop key="Timed Pressure" value="' + array_to_sv(p_v) + '" />\n')
                    f.write(t * 4 + '<prop key="Values" value="' + a_val + '" />\n')
                else:
                    raise ValueError('Boundary condition ' + tp + ' unknown')
    
                f.write(t * 3 + '</cap>\n')
    
            # inflow
            f.write(t * 3 + '<cap name="inflow">\n')
            print_props(f, inflow_props, t * 4)
            f.write(t * 3 + '</cap>\n')
    
            f.write(t * 2 + '</cap_props>\n')
    
            # wall
            f.write(t * 2 + '<wall_props>\n')
            print_props(f, wall_props, t * 3)
            f.write(t * 2 + '</wall_props>\n')
    
            # various
            f.write(t * 2 + '<var_props />\n')
    
            # solver
            f.write(t * 2 + '<solver_props>\n')
            print_props(f, solver_props, t * 3)
            f.write(t * 2 + '</solver_props>\n')
    
            # run
            f.write(t * 2 + '<run_props>\n')
            print_props(f, run_props, t * 3)
            f.write(t * 2 + '</run_props>\n')
    
            # close
            f.write(t + '</job>\n')
            f.write('</mitk_job>')
    
    def write_bc(self, fdir, write_face=True, model='3d'):
        # get boundary conditions
        bc_def = self.opt['bc']
    
        # check if bc-file exists
        if not bc_def:
            return None, 'boundary conditions do not exist'
    
        # get outlet names
        outlets = self.db.get_outlet_names(self.geo)
    
        # names expected by svsolver for different boundary conditions
        bc_file_names = {'rcr': 'rcrt.dat', 'resistance': 'resistance.dat', 'coronary': 'cort.dat'}
    
        # keyword to indicate a new boundary condition
        keywords = {'rcr': '2', 'coronary': '1001'}
    
        # create bc-files for every bc type
        u_bc_types = list(set(bc_def['bc_type'].values()))
        files = {}
        fnames = []
        for t in u_bc_types:
            if t in bc_file_names:
                fname = os.path.join(fdir, bc_file_names[t])
                files[t] = open(fname, 'w+')
                fnames += [fname]
    
                # write keyword for new faces in first line
                if t == 'rcr' or t == 'coronary':
                    files[t].write(keywords[t] + '\n')
            else:
                return None, 'boundary condition not implemented (' + t + ')'
    
        # write boundary conditions
        for s in outlets:
            bc = bc_def['bc'][s]
            t = bc_def['bc_type'][s]
            f = files[t]
            write_vals = lambda names: f.write('\n'.join([str(bc[v]) for v in names]))
            if t == 'rcr':
                f.write(keywords[t] + '\n')
                if write_face:
                    f.write(s + '\n')
                write_vals(['Rp', 'C', 'Rd'])
                if 'Po' in bc and bc['Po'] != 0.0:
                    p_ref = bc['Po']
                else:
                    p_ref = 0.0
                f.write('\n0.0 ' + str(p_ref) + '\n')
                f.write('1.0 ' + str(p_ref) + '\n')
            elif t == 'resistance':
                f.write(s + ' ')
                f.write(str(bc['R']) + ' ')
                f.write(str(bc['Po']) + '\n')
            elif t == 'coronary':
                f.write(keywords[t] + '\n')
                if model == '1d':
                    f.write(s + '\n')
                write_vals(['q0', 'q1', 'q2', 'p0', 'p1', 'p2', 'b0', 'b1', 'b2', 'dQinidT', 'dPinidT'])
                f.write('\n')
    
                # write time and pressure pairs
                for m in bc_def['coronary'][bc['Pim']]:
                    f.write(str(m[0]) + ' ' + str(m[1]) + '\n')
    
        # close all opened files
        for t in u_bc_types:
            files[t].close()
    
        return fnames, False
    
    def copy_files(self):
        # get solver options
        opt = self.opt
    
        # define paths
        sim_dir = self.db.get_solve_dir_3d(self.geo)
        fpath_surf = os.path.join(sim_dir, 'mesh-complete', 'mesh-surfaces')
    
        # create simulation folder
        os.makedirs(fpath_surf, exist_ok=True)
    
        # copy inflow
        np.savetxt(os.path.join(sim_dir, 'inflow.flow'), opt['inflow_data'])
        # shutil.copy(self.db.get_sv_flow_path(self.geo, '3d'), os.path.join(sim_dir, 'inflow.flow'))
    
        # copy cap meshes
        for f in glob.glob(os.path.join(self.db.get_sv_meshes(self.geo), 'caps', '*.vtp')):
            shutil.copy(f, fpath_surf)
    
        # copy surface and volume mesh
        shutil.copy(self.db.get_sv_surface(self.geo), os.path.join(sim_dir, opt['mesh_vtp']))
        shutil.copy(self.db.get_volume_mesh(self.geo), os.path.join(sim_dir, opt['mesh_vtu']))
    
        # copy initial condition mesh
        shutil.copy(self.db.get_initial_conditions(self.geo), os.path.join(sim_dir, opt['mesh_initial']))
    
        # copy wall mesh
        shutil.copy(os.path.join(self.db.get_sv_meshes(self.geo), 'walls_combined.vtp'), os.path.join(sim_dir, opt['mesh_walls']))
    
    def copy_file(self, src, trg_dir):
        trg = os.path.join(self.db.get_svproj_dir(self.geo), self.db.svproj.dir[trg_dir], os.path.basename(src))
        shutil.copy2(src, trg)
    
    def make_folders(self):
        # make all project sub-folders
        for s in self.db.svproj.dir.values():
            os.makedirs(os.path.join(self.db.get_svproj_dir(self.geo), s), exist_ok=True)
    
        # copy image
        if self.db.get_img(self.geo) is not None:
            self.copy_file(self.db.get_img(self.geo), 'images')
    
        # copy volume mesh
        self.copy_file(self.db.get_volume_mesh(self.geo), 'meshes')
    
        # copy surface mesh
        self.copy_file(self.db.get_sv_surface(self.geo), 'meshes')
        self.copy_file(self.db.get_sv_surface(self.geo), 'models')
    
        return True
    
    def check_files(self):
        # check if files exist
        if self.db.get_volume_mesh(self.geo) is None:
            raise RuntimeError('no volume mesh')
        if self.db.get_sv_surface(self.geo) is None:
            raise RuntimeError('no SV surface mesh')
        # if self.db.get_img(self.geo) is None:
        #     raise RuntimeError('no medical image')


def print_props(f, props, t):
    for h in props:
        f.write(t + '<prop key="' + h[0] + '" value="' + h[1] + '" />\n')


def array_to_sv(array):
    sv_str = ''
    for j, (t, i) in enumerate(array):
        sv_str += str(t) + ' ' + str(i)
        if j < array.shape[0] - 1:
            sv_str += '&#x0A;'
    return sv_str


def rc_to_r(bc_def):
    """
    Convert all boundary conditions to resistance (for steady flow simulation)
    """
    for s, t in bc_def['bc_type'].items():
        bc = bc_def['bc'][s]
        bc_def['bc_type'][s] = 'resistance'
        if t == 'resistance':
            pass
        elif t == 'rcr':
            bc_def['bc'][s] = {'R': bc['Rp'] + bc['Rd'], 'Po': bc['Po']}
        elif t == 'coronary':
            bc_sv = coronary_sv_to_oned(bc)
            bc_def['bc'][s] = {'R': bc_sv['Ra1'] + bc_sv['Ra2'] + bc_sv['Rv1'], 'Po': bc_sv['P_v']}
    return bc_def


def main(db, geometries, params):
    for geo in geometries:

        if params.mode is not None:
            mode = params.mode
        else:
            mode = ''

        print('Running geometry ' + geo)
        try:
            # get meshes
            get_meshes(db, geo)

            # create sv project
            pj = Project(db, geo, mode)
            pj.create_sv_project()
        except Exception as e:
            # print(e)
            continue


if __name__ == '__main__':
    descr = 'Generate an svproject folder'
    d, g, p = input_args(descr)
    main(d, g, p)



#!/usr/bin/env python

import pdb
import sys
import os
import vtk
import shutil

from collections import defaultdict
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
import numpy as np
from vtk.util.numpy_support import vtk_to_numpy as v2n
from vtk.util.numpy_support import numpy_to_vtk as n2v

from lift_laplace import StiffnessMatrix
from get_database import Database, SimVascular, Post, input_args
from vtk_functions import read_geo, write_geo, ClosestPoints, cell_connectivity, region_grow, collect_arrays
from simulation_io import collect_results, get_caps_db
from compare_1d import plot_1d_3d_interior as plt1d3d


def plot_projection(db, geometries):
    for geo in geometries:
        # plot options
        opt = {'legend_col': False,
               'legend_row': False,
               'sharex': True,
               'sharey': 'row',
               'dpi': 200,
               'w': 15,
               'h': 6}

        res = defaultdict(lambda: defaultdict(dict))
        time = {}

        time_inflow, _ = db.get_inflow_smooth(geo)

        f_asym = '/home/pfaller/work/osmsc/data_generated/asymptotic/' + geo + '.vtp'
        f_proj = '/home/pfaller/work/osmsc/data_generated/initial_pressure/' + geo + '.vtp'
        # f_res_1d = db.get_1d_flow_path(geo)
        f_res_1d = '/home/pfaller/work/osmsc/studies/1spb_length/1d_flow_last/' + geo + '.npy'
        f_rerun = '/home/pfaller/work/osmsc/studies/ini_1d_quad/3d_flow/' + geo + '.vtp'
        f_oned = db.get_1d_geo(geo)

        # collect_results('3d_rerun', res, time, f_asym, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo))
        collect_results('3d_rerun', res, time, f_rerun, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo))
        collect_results('3d', res, time, f_proj, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo))
        collect_results('1d', res, time, f_res_1d, f_oned, t_in=time_inflow[-1])

        plot_1d_3d_interior(db, opt, geo, res, time)

        # new plot
        res0 = defaultdict(lambda: defaultdict(dict))
        time0 = {}

        time_inflow, _ = db.get_inflow_smooth(geo)
        f_rerun = '/home/pfaller/work/osmsc/studies/ini_1d_quad/3d_flow/' + geo + '_first100.vtp'
        collect_results('3d_rerun', res0, time0, f_rerun, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo))

        res1 = defaultdict(lambda: defaultdict(dict))
        time1 = {}

        time_inflow, _ = db.get_inflow_smooth(geo)
        f_rerun = '/home/pfaller/work/osmsc/studies/ini_zero/3d_flow/' + geo + '_first1000.vtp'
        collect_results('3d_rerun', res1, time1, f_rerun, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo))

        res2 = defaultdict(lambda: defaultdict(dict))
        time2 = {}

        time_inflow, _ = db.get_inflow_smooth(geo)
        f_rerun = '/home/pfaller/work/osmsc/studies/ini_steady/3d_flow/' + geo + '_first100.vtp'
        collect_results('3d_rerun', res2, time2, f_rerun, t_in=time_inflow[-1], dt_3d=db.get_3d_timestep(geo))

        plot_1d_3d_interior_time(db, opt, geo, res0, time0, res1, res2)

def plot_1d_3d_interior(db, opt, geo, res, time):
    # get post-processing constants
    post = Post()

    # get models
    models = [k[:-4] for k in time.keys() if '_all' in k]

    # get 1d/3d map
    caps = get_caps_db(db, geo)
    cap_br = list(caps.values())
    cap_names = list(caps.keys())

    names = db.get_cap_names(geo)
    names['RCCA'] = 'Right common carotid'

    if len(res) > 50:
        dpi = opt['dpi'] // 4
        sharey = False
    else:
        dpi = opt['dpi']
        sharey = opt['sharey']

    fig, ax = plt.subplots(len(post.fields), len(res), figsize=(opt['w'], opt['h']), dpi=dpi, sharex='col', sharey=sharey)

    for i, f in enumerate(['pressure', 'flow']):
        for j, br in enumerate(res.keys()):
            ax[i, j].grid(True)

            if opt['legend_row'] or i == 0:
                if br in cap_br:
                    name = names[cap_names[cap_br.index(br)]]
                    # if not name.isupper():
                    #     name = name.capitalize()
                else:
                    name = 'branch ' + str(br)
                ax[i, j].set_title(name)
            if opt['legend_row'] or i == len(post.fields) - 1:
                ax[i, j].set_xlabel('Vessel path [-]')
                ax[i, j].xaxis.set_tick_params(which='both', labelbottom=True)
            if opt['legend_col'] or j == 0:
                ax[i, j].set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')
                ax[i, j].yaxis.set_tick_params(which='both', labelleft=True)

            lg = []
            for m in models:
                path = res[br][m + '_path']
                # if f == 'flow' and m == '1d':
                #     pdb.set_trace()
                if m == '3d_rerun':
                    ax[i, j].plot(path / path[-1], res[br][f][m + '_int_last'][:, -1] * post.convert[f], post.styles[m])
                else:
                    ax[i, j].plot(path / path[-1], res[br][f][m + '_int'][:, -1] * post.convert[f], post.styles[m])
                lg.append(m)

            # ax[i, j].legend(lg)
            ax[i, j].set_xlim(0, 1)
            ax[i, j].set_xticks([0, 1])
            if f == 'flow':
                ax[i, j].set_ylim(0, 2.5)

    # add_image(db, geo, fig)
    # fig.tight_layout(rect=[0, 0.03, 1, 0.95])
    fig.savefig(db.get_post_path(geo, 'interior'), bbox_inches='tight')
    plt.close(fig)

def plot_1d_3d_interior_time(db, opt, geo, res, time, res1, res2):
    # get post-processing constants
    post = Post()

    # get models
    models = [k[:-4] for k in time.keys() if '_all' in k]

    # get 1d/3d map
    caps = get_caps_db(db, geo)
    cap_br = list(caps.values())
    cap_names = list(caps.keys())

    names = db.get_cap_names(geo)
    names['RCCA'] = 'Right common carotid'

    if len(res) > 50:
        dpi = opt['dpi'] // 4
        sharey = False
    else:
        dpi = opt['dpi']
        sharey = opt['sharey']

    fig, ax = plt.subplots(1, len(res), figsize=(15, 3), dpi=dpi, sharey=sharey)

    f = 'flow'
    m = '3d_rerun'
    n_max = 100
    dt = db.get_3d_timestep(geo)
    for j, br in enumerate(res.keys()):
        ax[j].grid(True)
        if br in cap_br:
            name = names[cap_names[cap_br.index(br)]]
            # if not name.isupper():
            #     name = name.capitalize()
        else:
            name = 'branch ' + str(br)
        ax[j].set_title(name)
        ax[j].set_xlabel('Time [s]')
        ax[j].xaxis.set_tick_params(which='both', labelbottom=True)
        # ax[j].xaxis.set_ticks(np.linspace(0, n_max * dt, 4))
        ax[j].xaxis.set_major_formatter(FormatStrFormatter('%.2f'))
        if opt['legend_col'] or j == 0:
            ax[j].set_ylabel('Max. flow deviation from mean flow [-]')
            ax[j].yaxis.set_tick_params(which='both', labelleft=True)

        path = res[br][m + '_path']
        results = [res1, res]#, res2
        colors = ['k', plt.get_cmap("tab10")(1)]#, 'r'
        for col, res_sol in zip(colors, results):
            diff = []
            for n in range(n_max):
                sol = res_sol[br][f][m + '_int'][:, n]

                # exclude caps
                if br == 0:
                    sol = sol[1:]
                else:
                    sol = sol[:-1]

                # relative difference
                diff += [np.max(np.abs(sol / np.mean(sol) - 1))]
                # ax[j].plot(path[1:] / path[-1], sol, post.styles[m])
            ax[j].semilogy(np.arange(n_max) * dt, diff, post.styles[m], color=col)
        ax[j].set_xlim([-0.001, n_max * dt])
        # ax[j].set_ylim([1e-5, 1e-1])
    fig.savefig(db.get_post_path(geo, 'interior_time'), bbox_inches='tight')
    plt.close(fig)


def project_1d_3d_lift(f_1d, f_vol, f_out, field):
    # read volume mesh
    vol = read_geo(f_vol).GetOutput()
    points_vol = v2n(vol.GetPoints().GetData())
    cells = cell_connectivity(vol)

    # read 1d results
    oned = read_geo(f_1d).GetOutput()
    points_1d = v2n(oned.GetPoints().GetData())

    # get volume points closest to centerline
    cp_vol = ClosestPoints(vol)
    ids_vol = np.unique(cp_vol.search(points_1d))

    # get centerline points closest to selected volume points
    cp_1d = ClosestPoints(oned)
    ids_cent = cp_1d.search(points_vol[ids_vol])

    # visualize imprint of centerline in volume mesh
    imprint = np.zeros(vol.GetNumberOfPoints())
    imprint[ids_vol] = 1
    arr = n2v(imprint)
    arr.SetName('imprint')
    vol.GetPointData().AddArray(arr)

    # get 1d field field
    field_1d = v2n(oned.GetPointData().GetArray(field))

    # create laplace FEM stiffness matrix
    laplace = StiffnessMatrix(cells['tetra'], points_vol)

    # solve laplace equation (map desired field from 1d to 3d)
    field_3d = laplace.HarmonicLift(ids_vol, field_1d[ids_cent])

    # create output array
    arr = n2v(field_3d)
    arr.SetName(field)
    vol.GetPointData().AddArray(arr)

    # write to file
    write_geo(f_out, vol)


def get_1d_3d_map(f_1d, f_vol):
    # read geoemtries
    vol = read_geo(f_vol).GetOutput()
    oned = read_geo(f_1d).GetOutput()

    # get points
    points_vol = v2n(vol.GetPoints().GetData())
    points_1d = v2n(oned.GetPoints().GetData())

    # get volume points closest to centerline
    cp_vol = ClosestPoints(vol)
    seed_points = np.unique(cp_vol.search(points_1d))

    # map centerline points to selected volume points
    cp_1d = ClosestPoints(oned)
    seed_ids = np.array(cp_1d.search(points_vol[seed_points]))

    # call region growing algorithm
    ids, dist, rad = region_grow(vol, seed_points, seed_ids, n_max=999)

    # check 1d to 3d map
    assert np.max(ids) <= oned.GetNumberOfPoints() - 1, '1d-3d map non-conforming'

    return ids, dist, rad


def add_array(geo, name, array):
    arr = n2v(array)
    arr.SetName(name)
    geo.GetPointData().AddArray(arr)


def project_1d_3d_grow(f_1d, f_vol, f_wall, f_out):
    # read geometries
    vol = read_geo(f_vol).GetOutput()
    cent = read_geo(f_1d).GetOutput()
    wall = read_geo(f_wall).GetOutput()

    # get 1d -> 3d map
    map_ids, map_iter, map_rad = get_1d_3d_map(f_1d, f_vol)

    # get arrays
    arrays_cent = collect_arrays(cent.GetPointData())

    # map all centerline arrays to volume geometry
    for name, array in arrays_cent.items():
        add_array(vol, name, array[map_ids])

    # add mapping to volume mesh
    for name, array in zip(['MapIds', 'MapIters'], [map_ids, map_iter]):
        add_array(vol, name, array)

    # inverse map
    map_ids_inv = {}
    for i in np.unique(map_ids):
        map_ids_inv[i] = np.where(map_ids == i)

    # create radial coordinate [0, 1]
    rad = np.zeros(vol.GetNumberOfPoints())
    for i, ids in map_ids_inv.items():
        rad_max = np.max(map_rad[ids])
        if rad_max == 0:
            rad_max = np.max(map_rad)
        rad[ids] = map_rad[ids] / rad_max
    add_array(vol, 'rad', rad)

    # set points at wall to hard 1
    wall_ids = collect_arrays(wall.GetPointData())['GlobalNodeID'].astype(int) - 1
    rad[wall_ids] = 1

    # mean velocity
    names = ['flow', 'velocity']
    for n in names:
        for a in arrays_cent.keys():
            if n in a:
                u_mean = arrays_cent[a] / arrays_cent['CenterlineSectionArea']

                # parabolic velocity
                u_quad = 2 * u_mean[map_ids] * (1 - rad**2)

                # scale parabolic flow profile to preserve mean flow
                for i, ids in map_ids_inv.items():
                    u_mean_is = np.mean(u_quad[map_ids_inv[i]])
                    u_quad[ids] *= u_mean[i] / u_mean_is

                # parabolic velocity vector field
                velocity = np.outer(u_quad, np.ones(3)) * arrays_cent['CenterlineSectionNormal'][map_ids]

                # add to volume mesh
                if n == 'velocity':
                    aname = a
                elif n == 'flow':
                    aname = 'velocity'
                add_array(vol, aname, velocity)

    # write to file
    write_geo(f_out, vol)


def get_error(f_3d, f_1d, f_out):
    geo_3d = read_geo(f_3d).GetOutput()
    geo_1d = read_geo(f_1d).GetOutput()
    arrays_3d = collect_arrays(geo_3d.GetPointData())
    arrays_1d = collect_arrays(geo_1d.GetPointData())

    for m in arrays_1d.keys():
        if 'pressure' in m:
            norm = np.mean(arrays_3d[m])
            err = np.abs(arrays_3d[m] - arrays_1d[m]) / norm
            add_array(geo_1d, 'error_' + m, err)
        if 'velocity' in m:
            norm = np.mean(np.linalg.norm(arrays_3d[m], axis=1))
            err = np.linalg.norm(arrays_3d[m] - arrays_1d[m], axis=1) / norm
            add_array(geo_1d, 'error_' + m, err)
            pdb.set_trace()
    write_geo(f_out, geo_1d)


def main(db, geometries):
    for geo in geometries:
        print(geo)
        f_vol = os.path.join(db.get_sv_meshes(geo), geo + '.vtu')
        f_0d = db.get_0d_flow_path_vtp(geo)
        f_1d = db.get_1d_flow_path_vtp(geo)
        f_wall = db.get_surfaces(geo, 'wall')
        f_out = db.get_initial_conditions_pressure(geo) #'test.vtu'#
        pdb.set_trace()

        if os.path.exists(f_out):
            print('  projection exists, skipping')
            continue

        if os.path.exists(f_0d):
            print('  using 0d')
            f_red = f_0d
        elif os.path.exists(f_1d):
            print('  using 1d')
            f_red = f_1d
        else:
            print('  no 0d/1d solution found')
            continue

        # if os.path.exists(f_out):
        #     print('  map exists')
        #     continue

        project_1d_3d_grow(f_red, f_vol, f_wall, f_out)


def main_paper():
    db = Database('deformable')
    geo = '0069_0001'
    f_vol = os.path.join(db.get_sv_meshes(geo), geo + '.vtu')
    f_red = db.get_0d_flow_path_vtp(geo)
    f_wall = db.get_surfaces(geo, 'wall')

    for m in ['0d', '1d']:
        if m == '0d':
            f_red = db.get_0d_flow_path_vtp(geo)
        elif m == '1d':
            f_red = db.get_1d_flow_path_vtp(geo)
        f_out = f_red.replace('.vtp', '.vtu')
        project_1d_3d_grow(f_red, f_vol, f_wall, f_out)


def convert_time(db, geometries):
    for geo in geometries:
        f_vol = os.path.join(db.get_sv_meshes(geo), geo + '.vtu')
        f_res = db.get_volume(geo)
        f_red = db.get_3d_flow(geo)
        f_wall = db.get_surfaces(geo, 'wall')
        d_out = os.path.join('/home/pfaller/work/osmsc/extrapolation/', geo)
        f_out = os.path.join(d_out, geo + '_mapped.vtu')
        f_err = os.path.join(d_out, geo + '_error.vtu')

        os.makedirs(d_out, exist_ok=True)

        project_1d_3d_grow(f_red, f_vol, f_wall, f_out)

        shutil.copy(f_res, os.path.join(d_out, geo + '.vtu'))
        shutil.copy(f_red, d_out)

        get_error(f_res, f_out, f_err)


if __name__ == '__main__':
    descr = 'Get 3D-3D statistics'
    # d, g, _ = input_args(descr)
    # main(d, g)
    main_paper()
    # convert_time(d, g)
    # plot_projection(d, g)
    # f_vol = '/home/pfaller/downloads/0069_0001_post-interv (1)/0069_0001_post-interv/Meshes/0069_0001.vtu'
    # f_1d = d.get_1d_flow_path_vtp('0069_0001')
    # f_wall = '/home/pfaller/downloads/0069_0001_post-interv (1)/0069_0001_post-interv/Meshes/wall.vtp'
    # f_out = '/home/pfaller/0069_0001_interv_ic.vtu'
    # project_1d_3d_grow(f_1d, f_vol, f_wall, f_out)



#!/usr/bin/env python

import os
import vtk
import argparse
import pdb

import numpy as np
import matplotlib.pyplot as plt

from get_database import input_args, Database, Post
from get_mean_flow_3d import extract_results
from simulation_io import get_dict, get_caps_db, collect_results_db_3d_3d


def transfer_array(node_trg, node_src, name):
    """
    Transfer point data from volume mesh to surface mesh using GlobalNodeID
    Args:
        node_trg: target mesh
        node_src: source mesh
        name: array name to be transfered
    """
    # get global node ids in both meshes
    nd_id_trg = v2n(node_trg.GetPointData().GetArray('GlobalNodeID')).astype(int)
    nd_id_src = v2n(node_src.GetPointData().GetArray('GlobalNodeID')).astype(int)

    # map source mesh to target mesh
    mask = map_meshes(nd_id_src, nd_id_trg)

    # transfer array from volume mesh to surface mesh
    assert node_src.GetPointData().HasArray(name), 'Source mesh has no array ' + name
    res = v2n(node_src.GetPointData().GetArray(name))

    # create array in target mesh
    out_array = n2v(res[mask])
    out_array.SetName(name)
    node_trg.GetPointData().AddArray(out_array)


def plot_bc(db, geo):
    # get post-processing constants
    post = Post()

    # get results
    res, time = collect_results_db_3d_3d(db, geo)

    # get caps
    caps = get_caps_db(db, geo)

    for i, f in enumerate(['flow', 'pressure']):
        names = []
        values = []
        for c, br in caps.items():
            values += [res[br][f]['3d_rerun_all'].tolist()]
            names += [c]

        fig, ax = plt.subplots(figsize=(16, 8), dpi=200)
        ax.plot(time['3d_rerun_all'], np.array(values).T * post.convert[f])
        ax.set_xlabel('Time [s]')
        ax.set_ylabel(f.capitalize() + ' [' + post.units[f] + ']')
        ax.legend(names)
        ax.grid(True)

        fpath = os.path.join(db.fpath_study, 'simulation', geo + '_' + f + '.png')
        fig.savefig(fpath, bbox_inches='tight')
        plt.close(fig)

def main(db, geometries):
    """
    Loop all geometries in database
    """
    for geo in geometries:
        fpath_1d = db.get_centerline_path(geo)
        fpath_vol = db.get_res_3d_vol_rerun(geo)
        fpath_out = db.get_3d_flow_rerun(geo)
        if not os.path.exists(fpath_vol):
            continue

        print('Processing ' + geo)
        extract_results(fpath_1d, fpath_vol, fpath_out, only_caps=True)
        plot_bc(db, geo)


if __name__ == '__main__':
    descr = 'Plot comparison of xd-results'
    d, g, _ = input_args(descr)
    main(d, g)



#!/usr/bin/env python

import os
import sys
import pdb
import tempfile
import shutil

from get_database import Database, input_args

# sys.path.append('/home/pfaller/work/repos/SimVascular_fork/Python/site-packages/')
sys.path.append('/home/pfaller/work/repos/SimVascular_zasexton/Python/site-packages/')

from sv_rom_simulation import centerlines


# list of geometries whos mesh is so coarse the vmtk centerline extraction failes
geos_coarse = ['0119_0001', '0085_1001', '0084_0001']
# completed: '0087_1001', '0005_1001',
# no fine surface mesh: '0082_1001'

def add_suffix(fpath, suffix):
    ext = os.path.splitext(fpath)
    return ext[0] + suffix + ext[1]


class Params(object):
    """
    Minimal parameter set for Centerlines class
    """
    def __init__(self, p):
        suffix = '_resampled'
        self.boundary_surfaces_dir = p['f_surf_caps']
        self.inlet_face_input_file = p['f_inflow']
        self.surface_model = p['f_surf_in']
        self.output_directory = os.path.dirname(add_suffix(p['f_outlet'], suffix))
        self.CENTERLINES_OUTLET_FILE_NAME = os.path.basename(add_suffix(p['f_outlet'], suffix))
        self.centerlines_output_file = add_suffix(p['f_cent_out'], suffix)
        self.cent_out = add_suffix(p['f_cent_out'], suffix)


def main(db, geometries):
    for geo in geometries:
        if not db.get_surfaces(geo, 'all_exterior'):
            continue

        if geo in geos_coarse:
            f_surf_in = os.path.join('/home/pfaller/work/osmsc/data_generated/mesh_fine', geo + '_fine.vtp')
        else:
            f_surf_in = db.get_surfaces(geo, 'all_exterior')

        # get model paths
        params = {'f_surf_in': f_surf_in,
                  'f_surf_caps': tempfile.mkdtemp(),
                  'f_inflow': os.path.basename(db.get_surfaces(geo, 'inflow')),
                  'f_outlet': db.get_centerline_outlet_path(geo),
                  'f_cent_out_vmtk': db.get_centerline_vmtk_path(geo),
                  'f_cent_out': db.get_centerline_path(geo),
                  'f_surf_out': db.get_surfaces_grouped_path(geo),
                  'f_sections_out': db.get_section_path(geo)}

        # if os.path.exists(params['f_cent_out']):
        #     continue
        print('Running geometry ' + geo)

        # copy cap surfaces to temp folder
        for f in db.get_surfaces(geo, 'caps'):
            shutil.copy2(f, params['f_surf_caps'])

        params = Params(params)

        # call SimVascular centerline extraction
        # try:
        cl = centerlines.Centerlines()
        cl.extract_center_lines(params)
        cl.write_outlet_face_names(params)
        # except Exception as e:
        #     print(e)


if __name__ == '__main__':
    descr = 'Generate a new surface mesh'
    d, g, _ = input_args(descr)
    main(d, g)



# Optimization for Vessel Connections

import numpy as np
from geomdl import BSpline, utilities
from geomdl.visualization import VisMPL
import numba as nb
from scipy import interpolate, optimize
import pyvista as pv
from tqdm import tqdm,trange
from time import perf_counter
from copy import deepcopy
# Function Factory for BSpline Generation
def bezier_factory(p1, p2, p3, p4, clamp_first=True, clamp_second=True):
    def create_bezier(data, p1_=p1, p2_=p2, p3_=p3, p4_=p4, c1=clamp_first, c2=clamp_second):
        v1 = p2_ - p1_
        v2 = p4_ - p3_
        l1 = np.linalg.norm(v1)
        l2 = np.linalg.norm(v2)
        v1 = v1 / l1
        v2 = v2 / l2
        if c1 and c2:
            if data[0] < 0:
                data[0] = 0
            if data[1] < 0:
                data[1] = 0
            if data[0] > 1:
                data[0] = 1
            if data[1] > 1:
                data[1] = 1
            fill_number = data.shape[0] - 2
            ctr_number = data.shape[0] - 2 + 18
            data_offset = 2
            ctr0 = p1_
            ctr1 = p1_ + v1 * (l1 / 2)
            ctr2 = p1_ + (1 + 3 * data[0]) * v1 * (l1 / 4)
            ctr3 = p3_ + (1 + 3 * data[1]) * v2 * (l2 / 4)
            ctr4 = p3_ + v2 * (l2 / 2)
            ctr5 = p3_
            ctr = np.zeros(ctr_number)
            fill_ctr = np.zeros(fill_number).reshape(-1, 3)
            t = np.linspace(0.25, 0.75, fill_ctr.shape[0])
            for idx in range(fill_ctr.shape[0]):
                fill_ctr[idx, :] = ctr2 * (1 - t[idx]) + ctr3 * (t[idx])
            fill_ctr = fill_ctr.flatten()
            ctr[0:3] = ctr0
            ctr[3:6] = ctr1
            ctr[6:9] = ctr2
            ctr[-9:-6] = ctr3
            ctr[-6:-3] = ctr4
            ctr[-3:] = ctr5
            ctr[9:-9] += fill_ctr + data[data_offset:]
        elif c1 and not c2:
            if data[0] < 0:
                data[0] = 0
            if data[0] > 1:
                data[1] = 1
            fill_number = data.shape[0] - 1
            ctr_number = data.shape[0] - 1 + 9
            data_offset = 1
            ctr1 = p1_
            ctr2 = p1_ + (1 + 3 * data[0]) * v1 * (l1 / 4)
            ctr3 = p3_
            ctr = np.zeros(ctr_number)
            fill_ctr = np.zeros(fill_number).reshape(-1, 3)
            t = np.linspace(0.1, 0.9, fill_ctr.shape[0])
            for idx in range(fill_ctr.shape[0]):
                fill_ctr[idx, :] = ctr2 * (1 - t[idx]) + ctr3 * (t[idx])
            fill_ctr = fill_ctr.flatten()
            ctr[0:3] = ctr1
            ctr[3:6] = ctr2
            ctr[-3:] = ctr3
            ctr[6:-3] += fill_ctr + data[data_offset:]
        elif not c1 and c2:
            if data[0] < 0:
                data[0] = 0
            if data[0] > 1:
                data[1] = 1
            fill_number = data.shape[0] - 1
            ctr_number = data.shape[0] - 1 + 9
            data_offset = 1
            ctr1 = p1_
            ctr2 = p3_ + (1 + 3 * data[2]) * v2 * (l2 / 4)
            ctr3 = p3_
            ctr = np.zeros(ctr_number)
            fill_ctr = np.zeros(fill_number).reshape(-1, 3)
            t = np.linspace(0.1, 0.9, fill_ctr.shape[0])
            for idx in range(fill_ctr.shape[0]):
                fill_ctr[idx, :] = ctr2 * (1 - t[idx]) + ctr3 * (t[idx])
            fill_ctr = fill_ctr.flatten()
            ctr[0:3] = ctr1
            ctr[-6:-3] = ctr2
            ctr[-3:] = ctr3
            ctr[6:-3] += fill_ctr + data[data_offset:]
        else:
            fill_number = data.shape[0]
            ctr_number = data.shape[0] + 6
            data_offset = 0
            ctr1 = p1_
            ctr2 = p3_
            ctr = np.zeros(ctr_number)
            fill_ctr = np.zeros(fill_number).reshape(-1, 3)
            t = np.linspace(0.1, 0.9, fill_ctr.shape[0])
            for idx in range(fill_ctr.shape[0]):
                fill_ctr[idx, :] = ctr1 * (1 - t[idx]) + ctr2 * (t[idx])
            fill_ctr = fill_ctr.flatten()
            ctr[0:3] = ctr1
            ctr[-3:] = ctr2
            ctr[3:-3] += fill_ctr + data[data_offset:]
        ctr = ctr.reshape(-1, 3).tolist()
        curve = BSpline.Curve()
        curve.degree = len(ctr) - 1
        curve.ctrlpts = ctr
        curve.knotvector = utilities.generate_knot_vector(curve.degree, len(curve.ctrlpts))
        curve.vis = VisMPL.VisCurve3D()
        return curve
    return create_bezier


# Code for calculating angles
@nb.jit(nopython=True)
def get_angle(v1,v2):
    return np.arccos(np.dot(-v1,v2))*(180/np.pi)

@nb.jit(nopython=True)
def get_all_angles(vectors):
    angles = np.zeros(vectors.shape[0]-1)
    for i in range(vectors.shape[0]-1):
        angles[i] = get_angle(vectors[i,:],vectors[i+1,:])
    return angles

def get_all_vectors(pts):
    vectors = pts[1:] - pts[:-1]
    vectors = vectors/np.linalg.norm(vectors, axis=1).reshape(-1, 1)
    return vectors

def cost_angles(angles, func=lambda x:-x*(x<0)):
    return np.sum(func(angles))

def cost_bounds(bounds, pts, func=lambda x:max(0,x)):
    bounds_score = 0.0
    for i in range(pts.shape[0]):
        bounds_score += func(bounds[0, 0] - pts[i, 0])
        bounds_score += func(pts[i, 0] - bounds[0, 1])
        bounds_score += func(bounds[1, 0] - pts[i, 1])
        bounds_score += func(pts[i, 1] - bounds[1, 1])
        bounds_score += func(bounds[2, 0] - pts[i, 2])
        bounds_score += func(pts[i, 2] - bounds[2, 1])
    return bounds_score

@nb.jit(nopython=True)
def close_exact(data, point, radius_buffer):
    line_direction = np.zeros((data.shape[0], 3))
    ss = np.zeros(data.shape[0])
    tt = np.zeros(data.shape[0])
    hh = np.zeros(data.shape[0])
    cc = np.zeros((data.shape[0], 3))
    cd = np.zeros(data.shape[0])
    line_distances = np.zeros(data.shape[0])
    for i in range(data.shape[0]):
        line_direction[i, :] = (data[i, 3:6] - data[i, 0:3])/np.linalg.norm(data[i, 3:6] - data[i, 0:3])
        ss[i] = np.dot(data[i, 0:3]-point, line_direction[i, :])
        tt[i] = np.dot(point-data[i, 3:6], line_direction[i, :])
        d = np.array([ss[i], tt[i], 0])
        hh[i] = np.max(d)
        diff = point - data[i, 0:3]
        cc[i, :] = np.cross(diff, line_direction[i, :])
        cd[i] = np.linalg.norm(cc[i, :])
        line_distances[i] = np.sqrt(hh[i]**2+cd[i]**2) - data[i, 6] - radius_buffer
    return line_distances

@nb.jit(nopython=True)
def cost_collisions_fast(collision_vessels, collision_centers, collision_lengths, r, pts, radius_buffer):
    collisions = 0
    for i in range(pts.shape[0]):
        tmp = np.argwhere((np.linalg.norm(collision_centers-pts[i,:])-collision_lengths-r-collision_vessels[:,6])<0).flatten()
        dist = close_exact(collision_vessels[tmp,:], pts[i, :], radius_buffer)
        collisions += np.sum((r+radius_buffer-dist)*((r+radius_buffer-dist)>0))
    return collisions

@nb.jit(nopython=True)
def cost_collisions(collision_vessels, r, pts, radius_buffer):
    collisions = 0
    #tmp_c = (collision_vessels[:,0:3] + collision_vessels[:,3:6])/2
    #tmp_l = np.linalg.norm(collision_vessels[:,0:3].T-collision_vessels[:,3:6].T)
    for i in range(pts.shape[0]):
        #tmp = np.argwhere(np.linalg.norm(tmp_c-pts[i,:])-tmp_l-r-collision_vessels[:,6]<0).flatten()
        dist = close_exact(collision_vessels, pts[i, :], radius_buffer)
        collisions += np.sum((r+radius_buffer-dist)*((r+radius_buffer-dist)>0))
    return collisions

def cost(data,curve_generator=None,r=None,p1=None,p2=None,p3=None,p4=None,collision_vessels=None,
         radius_buffer=None,bounds=None,sample_size=None,clamp_first=None,
         clamp_second=None,length_threshold=1.1,angle_threshold=110,collision_centers=None,
         collision_lengths=None):
    curve = curve_generator(data, c1=clamp_first, c2=clamp_second)
    curve.sample_size = sample_size
    curve.evaluate()
    pts = np.array(curve.evalpts)
    cpts = pts[2:-2,:]
    if collision_vessels is not None:
        collision_cost = cost_collisions_fast(collision_vessels, collision_centers, collision_lengths, r, cpts, radius_buffer)
        #collision_cost = cost_collisions(collision_vessels, r, cpts, radius_buffer)
    else:
        collision_cost = 0
    bounds_cost = cost_bounds(bounds, pts)
    vectors = get_all_vectors(pts)
    angles = get_all_angles(vectors)
    angle_cost = cost_angles(angles-angle_threshold)
    spline_length = np.sum(np.linalg.norm(np.diff(pts,axis=0),axis=1))
    conn_length = np.linalg.norm(p2-p4)
    vessel_1_length = np.linalg.norm(p2-p1)
    vessel_2_length = np.linalg.norm(p4-p3)
    spline_length = spline_length/(conn_length+vessel_1_length+vessel_2_length)
    return (bounds_cost+angle_cost+collision_cost+max(spline_length-length_threshold,0))*spline_length

class connection:
    def __init__(self):
        self.P1 = None
        self.P2 = None
        self.P3 = None
        self.P4 = None
        self.R1 = None
        self.R2 = None
        self.bounds = None
        self.xopt = None
        self.plotting_vessels = None
    def set_solver(self, method='basinhopping'):
        if method == 'basinhopping':
            self.solver = optimize.basinhopping
        elif method == 'shgo':
            self.solver = optimize.shgo
        elif method == 'differential_evolution':
            self.solver = optimize.differential_evolution
        else:
            print('Solver method not implemented')
        return
    def set_vessel(self,id,p1,p2,r):
        if id == 1:
            self.P1 = p1
            self.P2 = p2
            self.R1 = r
        elif id == 2:
            self.P3 = p1
            self.P4 = p2
            self.R2 = r
        return
    def set_bounds(self,bounds):
        self.bounds = bounds
        return
    def set_collision_vessels(self,collision_vessels):
        self.plotting_vessels = collision_vessels
        filter_1_x = np.argwhere(collision_vessels[:, 0] != self.P1[0]).flatten().tolist()
        filter_1_y = np.argwhere(collision_vessels[:, 1] != self.P1[1]).flatten().tolist()
        filter_1_z = np.argwhere(collision_vessels[:, 2] != self.P1[2]).flatten().tolist()
        filter_1 = []
        filter_1.extend(filter_1_x)
        filter_1.extend(filter_1_y)
        filter_1.extend(filter_1_z)
        filter_1 = np.array(list(set(filter_1)))
        collision_vessels = collision_vessels[filter_1,:]
        filter_2_x = np.argwhere(collision_vessels[:, 0] != self.P3[0]).flatten().tolist()
        filter_2_y = np.argwhere(collision_vessels[:, 1] != self.P3[1]).flatten().tolist()
        filter_2_z = np.argwhere(collision_vessels[:, 2] != self.P3[2]).flatten().tolist()
        filter_2 = []
        filter_2.extend(filter_2_x)
        filter_2.extend(filter_2_y)
        filter_2.extend(filter_2_z)
        filter_2 = np.array(list(set(filter_2)))
        collision_vessels = collision_vessels[filter_2,:]
        filter_3_x = np.argwhere(collision_vessels[:, 3] != self.P1[0]).flatten().tolist()
        filter_3_y = np.argwhere(collision_vessels[:, 4] != self.P1[1]).flatten().tolist()
        filter_3_z = np.argwhere(collision_vessels[:, 5] != self.P1[2]).flatten().tolist()
        filter_3 = []
        filter_3.extend(filter_3_x)
        filter_3.extend(filter_3_y)
        filter_3.extend(filter_3_z)
        filter_3 = np.array(list(set(filter_3)))
        collision_vessels = collision_vessels[filter_3,:]
        filter_4_x = np.argwhere(collision_vessels[:, 3] != self.P3[0]).flatten().tolist()
        filter_4_y = np.argwhere(collision_vessels[:, 4] != self.P3[1]).flatten().tolist()
        filter_4_z = np.argwhere(collision_vessels[:, 5] != self.P3[2]).flatten().tolist()
        filter_4 = []
        filter_4.extend(filter_4_x)
        filter_4.extend(filter_4_y)
        filter_4.extend(filter_4_z)
        filter_4 = np.array(list(set(filter_4)))
        collision_vessels = collision_vessels[filter_4,:]
        self.collision_vessels = collision_vessels
        self.collision_centers = (collision_vessels[:,0:3]+collision_vessels[:,3:6])/2
        self.collision_lengths = np.linalg.norm(collision_vessels[:,3:6]-collision_vessels[:,0:3],axis=1)
        return
    def solve(self,*args,clamp_first=True,clamp_second=True,sample_size=20,radius_buffer=0.01):
        start = perf_counter()
        if len(args) == 0:
            number_free_points = 2
        else:
            number_free_points = args[0]
        self.radius_buffer = radius_buffer
        self.generator = bezier_factory(self.P1,self.P2,self.P3,self.P4,clamp_first=clamp_first,clamp_second=clamp_second)
        cost_function = lambda data: cost(data,curve_generator=self.generator,r=self.R1,p1=self.P1,p2=self.P2,
                                          p3=self.P3,p4=self.P4,
                                          collision_vessels=self.collision_vessels,radius_buffer=radius_buffer,
                                          sample_size=sample_size,bounds=self.bounds,clamp_first=clamp_first,
                                          clamp_second=clamp_second,collision_centers=self.collision_centers,
                                          collision_lengths=self.collision_lengths)
        if clamp_first and clamp_second:
            clamp_number = 2
        elif clamp_first or clamp_second:
            clamp_number = 1
        else:
            clamp_number = 0
        #if self.xopt is None:
        self.xopt = np.zeros(clamp_number+3*number_free_points)
        #if not self.xopt.shape[0] == clamp_number+3*number_free_points:
        #self.xopt = np.zeros(clamp_number+3*number_free_points)
        shortest_length = np.linalg.norm(self.P2-self.P4)
        lb = np.ones(self.xopt.shape[0])*(-shortest_length/number_free_points)
        for i in range(clamp_number):
            lb[i] = 0
            self.xopt[i] = 1
        ub = np.ones(self.xopt.shape[0])*(shortest_length/number_free_points)
        for i in range(clamp_number):
            ub[i] = 1
        solver_bounds = []
        for b in range(len(lb)):
            solver_bounds.append([lb[b],ub[b]])
        res = self.solver(cost_function, self.xopt)#,bounds=solver_bounds)
        self.xopt = res.x
        self.best = res.fun
        self.curve = self.generator(self.xopt)
        self.curve.sample_size = sample_size
        self.curve.evaluate()
        res.elapsed_time = perf_counter() - start
        return res
    def build_vessels(self,seperate):
        pts = np.array(self.curve.evalpts)
        if seperate:
            sep = (pts.shape[0]-1)//2
            vessels_1 = np.zeros((sep,7))
            vessels_2 = np.zeros((pts.shape[0]-1-sep,7))
            vessels_1_pts = pts[:sep+1,:]
            vessels_2_pts = np.flip(pts[sep:,:],axis=0)
            vessels_1[:,0:3] = vessels_1_pts[:-1,:]
            vessels_1[:,3:6] = vessels_1_pts[1:,:]
            vessels_2[:,0:3] = vessels_2_pts[:-1,:]
            vessels_2[:,3:6] = vessels_2_pts[1:,:]
            vessels_1[:,6] = self.R1
            vessels_2[:,6] = self.R2
        else:
            vessels_1 = np.zeros((pts.shape[0]-1,7))
            vessels_1[:,0:3] = pts[:-1,:]
            vessels_1[:,3:6] = pts[1:,:]
            vessels_1[:,6] = self.R1
            pts_reverse = np.flip(pts,axis=0)
            vessels_2 = np.zeros((pts.shape[0]-1,7))
            vessels_2[:,0:3] = pts_reverse[:-1,:]
            vessels_2[:,3:6] = pts_reverse[1:,:]
            vessels_2[:,6] = self.R2
        self.vessels_1 = vessels_1
        self.vessels_2 = vessels_2
        self.seperate = seperate
        return vessels_1,vessels_2
    def show(self):
        plotter = pv.Plotter()
        if self.seperate:
            for i in range(self.vessels_1.shape[0]):
                center = (self.vessels_1[i,0:3]+self.vessels_1[i,3:6])/2
                direction = self.vessels_1[i,3:6] - self.vessels_1[i,0:3]
                length = np.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_1[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='red')
            for i in range(self.vessels_2.shape[0]):
                center = (self.vessels_2[i,0:3]+self.vessels_2[i,3:6])/2
                direction = self.vessels_2[i,3:6] - self.vessels_2[i,0:3]
                length = np.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_2[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='blue')
        else:
            for i in range(self.vessels_1.shape[0]):
                center = (self.vessels_1[i,0:3]+self.vessels_1[i,3:6])/2
                direction = self.vessels_1[i,3:6] - self.vessels_1[i,0:3]
                length = np.linalg.norm(direction)
                direction = direction/length
                radius = self.vessels_1[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='red')
        if self.plotting_vessels is not None:
            for i in range(self.plotting_vessels.shape[0]):
                center = (self.plotting_vessels[i,0:3]+self.plotting_vessels[i,3:6])/2
                direction = self.plotting_vessels[i,3:6] - self.plotting_vessels[i,0:3]
                length = np.linalg.norm(direction)
                direction = direction/length
                radius = self.plotting_vessels[i,6]
                cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
                plotter.add_mesh(cylinder,color='black')
        center = (self.P1+self.P2)/2
        direction = (self.P2-self.P1)
        length = np.linalg.norm(direction)
        direction = direction/length
        radius = self.R1
        cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
        plotter.add_mesh(cylinder,color='yellow')
        center = (self.P3+self.P4)/2
        direction = (self.P4-self.P3)
        length = np.linalg.norm(direction)
        direction = direction/length
        radius = self.R1
        cylinder = pv.Cylinder(center=center,direction=direction,radius=radius,height=length)
        plotter.add_mesh(cylinder,color='yellow')
        return plotter
    def check_bounds(self):
        start = perf_counter()
        pts = np.array(self.curve.evalpts)
        cost_result = cost_bounds(self.bounds,pts)
        print(perf_counter()-start)
        return cost_result
    def check_angles(self):
        start = perf_counter()
        pts = np.array(self.curve.evalpts)
        vectors = get_all_vectors(pts)
        angles = get_all_angles(vectors)
        cost_result = cost_angles(angles)
        print(perf_counter()-start)
        return cost_result
    def check_collisions(self):
        start = perf_counter()
        pts = np.array(self.curve.evalpts)
        pts = pts[2:-2,:]
        if self.collision_vessels is not None:
            collision_cost = cost_collisions_fast(self.collision_vessels, self.collision_centers,self.collision_lengths,self.R1, pts, self.radius_buffer)
            #collision_cost = cost_collisions(self.collision_vessels, self.R1, pts, self.radius_buffer)
        else:
            collision_cost = 0
        print(perf_counter()-start)
        return collision_cost

class simple_connection:
    def __init__(self,forest,network,tree_1,tree_2,connection_id,radius_buffer,
                 clamp_first,clamp_second,seperate):
        self.conn = connection()
        self.forest = forest
        self.network = network
        self.tree_1 = tree_1
        self.tree_2 = tree_2
        self.vessel_1_id = forest.assignments[network][tree_1][connection_id]
        self.vessel_2_id = forest.assignments[network][tree_2][connection_id]
        vessel_1 = forest.networks[network][tree_1].data[self.vessel_1_id]
        vessel_2 = forest.networks[network][tree_2].data[self.vessel_2_id]
        self.conn.set_vessel(1,vessel_1[0:3],vessel_1[3:6],vessel_1[21])
        self.conn.set_vessel(2,vessel_2[0:3],vessel_2[3:6],vessel_2[21])
        tree_1_ids = np.argwhere(forest.networks[network][tree_1].data[:, 15] > -1).flatten()
        tree_2_ids = np.argwhere(forest.networks[network][tree_2].data[:, 15] > -1).flatten()
        #tree_1_ids = np.delete(tree_1_ids,int(vessel_1[17]))
        #tree_2_ids = np.delete(tree_2_ids,int(vessel_2[17]))
        tree_1_data = forest.networks[network][tree_1].data[tree_1_ids, :]
        tree_2_data = forest.networks[network][tree_2].data[tree_2_ids, :]
        collision_vessels = np.zeros((tree_1_data.shape[0]+tree_2_data.shape[0],7))
        collision_vessels[:tree_1_data.shape[0],0:3] = tree_1_data[:,0:3]
        collision_vessels[:tree_1_data.shape[0],3:6] = tree_1_data[:,3:6]
        collision_vessels[:tree_1_data.shape[0],6] = tree_1_data[:,21]
        collision_vessels[tree_1_data.shape[0]:,0:3] = tree_2_data[:,0:3]
        collision_vessels[tree_1_data.shape[0]:,3:6] = tree_2_data[:,3:6]
        collision_vessels[tree_1_data.shape[0]:,6] = tree_2_data[:,21]
        self.radius_buffer = radius_buffer
        self.clamp_first = clamp_first
        self.clamp_second = clamp_second
        self.conn.set_collision_vessels(collision_vessels)
        self.bounds = np.array([forest.boundary.x_range, forest.boundary.y_range, forest.boundary.z_range])
        self.seperate = seperate
        self.edge_1 = vessel_1
        self.edge_2 = vessel_2
    def set_solver(self,*args,method='basinhopping'):
        self.conn.set_solver(method=method)
        self.conn.set_bounds(self.bounds)
    def solve(self,*args):
        if len(args) > 0:
            number_ctl_points = args[0]
        else:
            number_ctl_points = 2
        self.res = self.conn.solve(number_ctl_points)
        return self.res
    def build(self):
        self.vessels_1,self.vessels_2 = self.conn.build_vessels(self.seperate)
    def show(self):
        plotter = self.conn.show()
        plotter.show()
    def export_vessels(self):
        if self.seperate:
            vessels_tree_1 = np.ones((self.vessels_1.shape[0],self.forest.networks[self.network][self.tree_1].data.shape[1]))*-1
            vessels_tree_2 = np.ones((self.vessels_2.shape[0],self.forest.networks[self.network][self.tree_2].data.shape[1]))*-1
            vessels_tree_1[:,0:3] = self.vessels_1[:,0:3]
            vessels_tree_1[:,3:6] = self.vessels_1[:,3:6]
            vessels_tree_2[:,0:3] = self.vessels_2[:,0:3]
            vessels_tree_2[:,3:6] = self.vessels_2[:,3:6]
            tree_1_dirs = vessels_tree_1[:,3:6] - vessels_tree_1[:,0:3]
            tree_2_dirs = vessels_tree_2[:,3:6] - vessels_tree_2[:,0:3]
            vessels_tree_1[:,20] = np.linalg.norm(tree_1_dirs,axis=1)
            vessels_tree_2[:,20] = np.linalg.norm(tree_2_dirs,axis=1)
            vessels_tree_1[:,12:15] = tree_1_dirs/vessels_tree_1[:,20].reshape(-1,1)
            vessels_tree_2[:,12:15] = tree_2_dirs/vessels_tree_2[:,20].reshape(-1,1)
            vessels_tree_1[:,21] = self.vessels_1[:,6]
            vessels_tree_2[:,21] = self.vessels_2[:,6]
            vessels_tree_1[:,22] = self.edge_1[22]
            vessels_tree_2[:,22] = self.edge_2[22]
            self.vessels = [vessels_tree_1,vessels_tree_2]
        else:
            print('not implemented yet')
class tree_connections:
    def __init__(self,forest,network,tree_1,tree_2,radius_buffer,clamp_first=True,clamp_second=True,seperate=True):
        self.forest = forest
        self.network = network
        self.tree_1 = tree_1
        self.tree_2 = tree_2
        self.radius_buffer = radius_buffer
        self.extra_collision_vessels = None
        self.connection_solutions = []
        for i in range(self.forest.assignments[self.network][self.tree_1].shape[0]):
            tmp_connection = simple_connection(self.forest,self.network,self.tree_1,self.tree_2,i,self.radius_buffer,clamp_first,clamp_second,seperate)
            self.connection_solutions.append(tmp_connection)
    def solve(self,clamp_first=True,clamp_second=True,seperate=True,max_ctrl_points=5):
        connection_solutions = []
        continue_optimize = []
        num_ctrl_points = 1
        for i in trange(self.forest.assignments[self.network][self.tree_1].shape[0]):
            tmp_connection = self.connection_solutions[i]
            tmp_connection.clamp_first = clamp_first
            tmp_connection.clamp_second = clamp_second
            tmp_connection.seperate = seperate
            tmp_connection.set_solver()
            for j in range(len(connection_solutions)):
                if connection_solutions[j].seperate:
                    collision_vessels = tmp_connection.conn.collision_vessels
                    collision_vessels = np.vstack((collision_vessels,connection_solutions[j].vessels_1,
                                                   connection_solutions[j].vessels_2))
                    tmp_connection.conn.set_collision_vessels(collision_vessels)
                else:
                    collision_vessels = tmp_connection.conn.collision_vessels
                    collision_vessels = np.vstack((collision_vessels,connection_solutions[j].vessels_1))
                    tmp_connection.conn.set_collision_vessels(collision_vessels)
            if self.extra_collision_vessels is not None:
                collision_vessels = tmp_connection.conn.collision_vessels
                collision_vessels = np.vstack((collision_vessels,self.extra_collision_vessels))
                tmp_connection.conn.set_collision_vessels(collision_vessels)
            tmp_connection.solve(num_ctrl_points)
            if tmp_connection.res.fun > 0:
                continue_optimize.append(i)
            tmp_connection.build()
            connection_solutions.append(tmp_connection)
        while len(continue_optimize) > 0 and num_ctrl_points <= max_ctrl_points:
            num_ctrl_points += 1
            new_continue_optimize = []
            print("Design space increased to {} control points for connections {}".format(num_ctrl_points,continue_optimize))
            while len(continue_optimize) > 0:
                i = continue_optimize.pop(0)
                solved_vessels = list(set(list(range(len(connection_solutions)))) - set(continue_optimize))
                tmp_connection = self.connection_solutions[i]
                for j in solved_vessels:
                    if connection_solutions[j].seperate:
                        collision_vessels = tmp_connection.conn.collision_vessels
                        collision_vessels = np.unique(np.vstack((collision_vessels,connection_solutions[j].vessels_1,
                                                       connection_solutions[j].vessels_2)),axis=0)
                        tmp_connection.conn.set_collision_vessels(collision_vessels)
                    else:
                        collision_vessels = tmp_connection.conn.collision_vessels
                        collision_vessels = np.unique(np.vstack((collision_vessels,connection_solutions[j].vessels_1)),axis=0)
                        tmp_connection.conn.set_collision_vessels(collision_vessels)
                if self.extra_collision_vessels is not None:
                    collision_vessels = tmp_connection.conn.collision_vessels
                    collision_vessels = np.vstack((collision_vessels, self.extra_collision_vessels))
                    tmp_connection.conn.set_collision_vessels(collision_vessels)
                tmp_connection.solve(num_ctrl_points)
                if tmp_connection.res.fun > 0:
                    new_continue_optimize.append(i)
                tmp_connection.build()
                if tmp_connection.res.fun < connection_solutions[i].res.fun:
                    connection_solutions[i] = tmp_connection
            continue_optimize = new_continue_optimize
        self.connection_solutions = connection_solutions
        return
    def add_collision_vessels(self,vessels):
        if self.extra_collision_vessels is not None:
            self.extra_collision_vessels = np.unique(np.vstack((self.extra_collision_vessels,vessels)),axis=0)
        else:
            self.extra_collision_vessels = vessels
    def build_vessels(self):
        for i in range(len(self.connection_solutions)):
            self.connection_solutions[i].export_vessels()

class network_connections:
    def __init__(self,forest,network,radius_buffer):
        self.forest = forest
        self.network = network
        self.radius_buffer = radius_buffer
    def solve(self,order=None):
        solutions = []
        if order is None:
            order = list(range(len(self.forest.trees_per_network[self.network])))
        for i in range(1,len(order)):
            if i == 1:
                tmp_tree_connections = tree_connections(self.forest,self.network,order[0],order[i],self.radius_buffer)
                for k in range(len(tmp_tree_connections.connection_solutions)):
                    tmp_tree_connections.connection_solutions[k].conn.P1 = solutions[0].connection_solutions[k].vessels_1[-1,0:3]
                    tmp_tree_connections.connection_solutions[k].conn.P2 = solutions[0].connection_solutions[k].vessels_1[-1,3:6]
                    tmp_tree_connections.connection_solutions[k].conn.R1 = solutions[0].connection_solutions[k].vessels_1[-1,6]
                remaining_trees = set(order)-set([order[0],order[i]])
                for j in remaining_trees:
                    vessels = np.zeros((self.forest.networks[self.network][j].data.shape[0],7))
                    vessels[:,0:3] = self.forest.networks[self.network][j].data[:,0:3]
                    vessels[:,3:6] = self.forest.networks[self.network][j].data[:,3:6]
                    vessels[:,6] = self.forest.networks[self.network][j].data[:,21]
                    tmp_tree_connections.add_collision_vessels(vessels)
                tmp_tree_connections.solve()
            else:
                solved_trees = list(range(1,i))
                remaining_trees = set(list(range(len(order)))) - set(solved_trees)
                tmp_tree_connections = tree_connections(self.forest,self.network,order[0],order[i],self.radius_buffer,
                                                        clamp_first=False,clamp_second=True,seperate=False)
                for k in range(len(tmp_tree_connections.connection_solutions)):
                    tmp_tree_connections.connection_solutions[k].conn.P1 = solutions[0].connection_solutions[k].vessels_1[-1,0:3]
                    tmp_tree_connections.connection_solutions[k].conn.P2 = solutions[0].connection_solutions[k].vessels_1[-1,3:6]
                    tmp_tree_connections.connection_solutions[k].conn.R1 = solutions[0].connection_solutions[k].vessels_1[-1,6]
                for j in solved_trees:
                    for conn in solutions[j].connection_solutions:
                        if conn.seperate:
                            vessels = np.zeros((conn.vessels_1.shape[0]+conn.vessels_2.shape[0],7))
                            vessels[0:conn.vessels_1.shape[0],:] = conn.vessels_1
                            vessels[conn.vessels_1.shape[0]:,:] = conn.vessels_2
                        else:
                            vessels = conn.vessels_1
                        tmp_tree_connections.add_collision_vessels(vessels)
                for j in remaining_trees:
                    vessels = np.zeros((self.forest.networks[self.network][j].data.shape[0],7))
                    vessels[:,0:3] = self.forest.networks[self.network][j].data[:,0:3]
                    vessels[:,3:6] = self.forest.networks[self.network][j].data[:,3:6]
                    vessels[:,6] = self.forest.networks[self.network][j].data[:,21]
                    tmp_tree_connections.add_collision_vessels(vessels)
                tmp_tree_connections.solve()
            solutions.append(tmp_tree_connections)

"""
pts = np.array([[-2,0,0],[-1,0,0],[2,0,0],[1,0,0]])
bounds = np.array([[-15,15],[-15,15],[-15,15]])
gen1 = bezier_factory(pts[0,:],pts[1,:],pts[2,:],pts[3,:])
gen2 = bezier_factory(pts[0,:],pts[1,:],pts[2,:],pts[3,:],clamp_first=False)
gen3 = bezier_factory(pts[0,:],pts[1,:],pts[2,:],pts[3,:],clamp_first=False,clamp_second=False)

c = connection()
c.set_vessel(1,pts[0,:],pts[1,:],0.1)
c.set_vessel(2,pts[2,:],pts[3,:],0.1)
c.set_solver()
c.set_bounds(bounds)
c.set_collision_vessels(None)

pts1 = 10*np.array([[0,0.25,0.5],[0.5,0.25,0.5],[0.25,1,0.5],[0.25,0.5,0.5]])
c1 = connection()
c1.set_vessel(1,pts1[0,:],pts1[1,:],0.1)
c1.set_vessel(2,pts1[2,:],pts1[3,:],0.1)
c1.set_solver()
c1.set_bounds(bounds)
collision_vessels = 15*np.random.random((15,7))
collision_vessels[:,6] = 0.05
c1.set_collision_vessels(collision_vessels)

pts2 = 10*np.array([[0.25,0,0],[0.5,0,0],[-0.25,0,0],[-0.5,0,0]])
c1 = connection()
c1.set_vessel(1,pts2[0,:],pts2[1,:],0.1)
c1.set_vessel(2,pts2[2,:],pts2[3,:],0.1)
c1.set_solver()
c1.set_bounds(bounds)
c1.set_collision_vessels(None)
"""


#!/usr/bin/env python

import os
import sys
import re
import vtk
import argparse
import pdb

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as mtick

from collections import defaultdict
from scipy.interpolate import interp1d

from vtk.util.numpy_support import numpy_to_vtk as n2v
from vtk.util.numpy_support import vtk_to_numpy as v2n

sys.path.append('..')
from get_database import Database, Post, input_args
from vtk_functions import read_geo, write_geo, get_all_arrays
from simulation_io import get_dict, get_caps_db, collect_results_db_3d_3d, collect_results_db_1d_3d, collect_results_db_0d
from get_sv_project import coronary_sv_to_oned
from get_statistics_bc import collect_errors

fsize = 12
plt.rcParams.update({'font.size': fsize})
# plt.style.use('dark_background')


def plot_correlation(db):
    post = Post()

    # get numerical time constants
    res_num = get_dict(db.get_convergence_path())

    # only plot subset of geometries
    geometries_paper = db.get_geometries_select('paper')

    # convergence criterion
    tol = 0.01

    fig1, ax1 = plt.subplots(1, 2, dpi=300, figsize=(12, 4), sharex='row')

    for pos, f in enumerate(['pressure', 'flow']):

        i_conv_all = []
        tau_all = []
        for geo in res_num:
            if geo not in geometries_paper:
                continue
            params = db.get_bcs(geo)

            # skip non-RCR models
            types = np.unique(list(params['bc_type'].values()))
            if not len(types) == 1 or not types[0] == 'rcr':
                continue

            i_conv = res_num[geo]['i_conv'][f]
            tau = np.mean(res_num[geo]['tau'][f])
            col = post.colors[params['params']['deliverable_category']]
            ax1[pos].plot(tau, i_conv, marker='o', color=col)
            i_conv_all += [i_conv]
            tau_all += [tau]

        # fit linear equation to data
        if db.study == '1spb_length' and f == 'pressure':
            # coef = np.polyfit(tau_all, i_conv_all, 1)
            # poly = np.poly1d(coef)
            # ax1.plot(tau_all, poly(tau_all), 'k-')

            t = np.linspace(0, 10, num=10000)
            fx = - np.log(tol) * t
            ax1[pos].plot(t, fx.astype(int) + 1, 'k-')

        # plt.yscale('log')
        ax1[pos].set_title(f.capitalize())
        ax1[pos].xaxis.grid('minor')
        ax1[pos].yaxis.grid(True)
        ax1[pos].set_xlim(0, 10)
        ax1[pos].set_ylim(0, 50)
        ax1[pos].set_xlabel(r'Model time constant $\bar{\tau}/T$ [-]')
        if pos == 0:
            ax1[pos].set_ylabel(r'Number of cardiac cycles [-]')

    fname = os.path.join(db.get_statistics_dir(), 'correlation.png')
    fig1.savefig(fname, bbox_inches='tight')
    plt.close(fig1)


def plot_convergence(db):
    post = Post()

    # get numerical time constants
    res_num = get_dict(db.get_convergence_path())

    fig1, ax1 = plt.subplots(dpi=400, figsize=(15, 6))

    markers = {'flow': 'o', 'pressure': 'x'}

    geos = []
    i = 0
    for geo in res_num:
        params = db.get_bcs(geo)
        i_conv = res_num[geo]['i_conv']
        col = post.colors[params['params']['deliverable_category']]
        for f, m in markers.items():
            ax1.plot(i, i_conv[f], marker=m, color=col)

        geos += [geo]
        i += 1

    # plt.yscale('log')
    ax1.xaxis.grid('minor')
    ax1.yaxis.grid(True)
    # ax1.set_ylim(0, 13)
    plt.xticks(np.arange(len(geos)), geos, rotation='vertical')
    plt.ylabel('Number of cardiac cycles [-]')
    fname = os.path.join(db.get_statistics_dir(), 'convergence.png')
    fig1.savefig(fname, bbox_inches='tight')


def make_err_plot(db, geo, ax, pos, m, f, p, res_m, errors, time, title_study=''):
    post = Post()
    t = 0

    # study names
    studies = {'ini_zero': 'Zero',
               'ini_steady': 'Steady',
               'ini_irene': 'Steady (start from mean)',
               'ini_1d_quad': '1D'}

    # get caps
    caps = get_caps_db(db, geo)
    del caps['inflow']

    # get boundary conditions
    bcs = db.get_bcs(geo)['bc']
    bct = db.get_bcs(geo)['bc_type']

    # get time
    n_cycle = time[m + '_n_cycle']

    # plot times
    times_plot = []
    times_all = []
    for k in range(1, n_cycle + 1):
        times_all += [np.where(time[m + '_i_cycle_' + str(k)])[0]]
        times_plot += [np.where(time[m + '_i_cycle_' + str(k)])[0][t]]
    times_all = np.array(times_all)

    # add last time step
    if t == 0:
        times_plot += [np.where(time[m + '_i_cycle_' + str(n_cycle)])[0][-1]]

    cycles = np.arange(len(times_plot))
    if t != 0:
        cycles += 1

    # error threshold for a converged solution
    thresh = 1.0e-2
    e_thresh = 'asymptotic'

    # get numerical time constant
    res_num = get_dict(db.get_convergence_path())
    if geo in res_num:
        tau = np.mean(res_num[geo]['tau']['pressure'])
        alpha = 1 / (np.exp(1 / tau) - 1)
        thresh_cyclic = thresh / alpha
    else:
        thresh_cyclic = np.nan
    # thresh_cyclic = 0
    # pdb.set_trace()
    #
    # collect results
    res_m_all = []
    res_m_t = []
    res_m_m = []
    res_0d_t = []
    res_qm_t = []
    for c, br in caps.items():
        res_m_all += [res_m[br][f][m + '_all']]
        res_m_t += [res_m[br][f][m + '_all'][times_plot]]
        # res_0d_t += [interp1d(res_0d[br]['t'], res_0d[br]['p'], fill_value='extrapolate')(time[m][t]).tolist()]
        res_0d_t += [res_m_t[-1][-1]]
        res_m_m += [np.mean(res_m[br][f][m + '_all'][times_all], axis=1)]

        if bct[c] == 'resistance':
            resistance = bcs[c]['R']
        elif bct[c] == 'rcr':
            resistance = bcs[c]['Rd'] + bcs[c]['Rp']
        elif bct[c] == 'coronary':
            cor = coronary_sv_to_oned(bcs[c])
            resistance = cor['Ra1'] + cor['Ra2'] + cor['Rv1']
        res_qm_t += [resistance * np.mean(res_m[br]['flow'][m + '_cap'])]

    # make plot
    c_max = 20
    x_min = 1
    style = 'x-'
    if p == 'cycle':
        title = 'Solution'
        x_min = 0
        style = '-'
        x = time[m + '_all'] / time[m][-1]
        y = np.array(res_m_all).T * post.convert[f]
        xticks = [0]
        ylabel = f.capitalize() + ' [' + post.units[f] + ']'
    elif p == 'cycle_norm':
        title = 'Normalized solution'
        x_min = 0
        xticks = [0]
        style = '-'
        x = time[m + '_all'] / time[m][-1]
        y = np.array(res_m_all).T / np.array(res_m_m)[:, -1]
        ylabel = f.capitalize() + ' [-]'
    elif p == 'initial':
        title = 'Initial values'
        x = cycles
        y = np.array(res_m_t).T * post.convert[f]
        xticks = [1]
        ylabel = 'Initial ' + f + ' [' + post.units[f] + ']'
    elif p == 'mean':
        title = 'Mean cycle solution'
        x = cycles[1:]
        y = np.array(res_m_m).T * post.convert[f]
        y /= y[-1]
        xticks = [1]
        # ylabel = 'Mean ' + f + ' [' + post.units[f] + ']'
        ylabel = 'Mean ' + f + ' [-]'
    elif p in 'cyclic':
        title = 'Cyclic error $\epsilon_n$'
        x = cycles[1:-1] + 1
        y = errors['cyclic'][f]
        xticks = [2]
        ylabel = 'Cyclic ' + f + ' error [-]'
    elif p in 'asymptotic':
        title = 'Asymptotic error $\epsilon_\infty$'
        x = cycles[1:-1]
        y = errors['asymptotic'][f][:-1]
        xticks = [1]
        ylabel = 'Asymptotic ' + f + ' error [-]'
    else:
        title = ''
        x = np.nan
        y = np.nan
        xticks = []
    xticks += [c_max]

    # converged time step
    conv = np.where(np.all(errors[e_thresh][f] < thresh, axis=1))[0]
    if not conv.any():
        i_conv = -1
    else:
        i_conv = np.min(conv)
    if e_thresh == 'cyclic':
        i_conv += 2
    elif e_thresh == 'asymptotic':
        i_conv += 1
        if p in 'asymptotic':
            print(f[0] + ' ' + str(i_conv))
    xticks += [i_conv]

    # plot
    if p == 'initial' or p == 'mean':
        plt.gca().set_prop_cycle(plt.rcParams['axes.prop_cycle'])
        # ax[pos].plot([cycles[0], cycles[-1]], np.vstack((y[-1], y[-1])), '--')
        ax[pos].plot([0, 999], [1, 1], 'k-')
        ax[pos].set_ylim([0, 1.2])
        # ax.plot([cycles[0], cycles[-1]], np.vstack((res_qm_t, res_qm_t)) * post.convert[f], '--')

    ax[pos].plot(x, y, style)
    ax[pos].axvline(x=i_conv, color='k')
    if not title_study or pos[1] == 0:
        ax[pos].set_ylabel(ylabel)
    x_eps = 0.5 * np.max(xticks) / 20
    ax[pos].set_xlim([x_min - x_eps, np.max(xticks) + x_eps])
    # ax[pos].set_xlim([0, np.max(x)])
    ax[pos].set_xticks(xticks)
    ax[pos].grid('both')

    if p in ['cyclic', 'asymptotic']:
        ax[pos].set_yscale('log')
        if f == 'pressure':
            ax[pos].set_ylim([1e-4, 1])
        elif f == 'flow':
            ax[pos].set_ylim([1e-5, 1e-1])
        # ax[pos].yaxis.set_major_formatter(mtick.PercentFormatter(1.0, 2))
    if p == 'cyclic' and e_thresh == 'asymptotic':
        ax[pos].plot([0, 999], [thresh_cyclic, thresh_cyclic], 'k-')
    if p == e_thresh:
        ax[pos].plot([0, 999], [thresh, thresh], 'k-')
    if pos[0] == 1:
        ax[pos].set_xlabel('Cardiac cycle [-]')
    if pos[0] == 0:
        if title_study:
            ax[pos].set_title(studies[title_study])
        else:
            ax[pos].set_title(title)
    ax[pos].set_prop_cycle(plt.rcParams['axes.prop_cycle'])

    return y, i_conv


def get_time_constants(db, geo, c_res):
    const = defaultdict(dict)

    # get cap names
    tau_ana_dic = db.get_time_constants(geo)
    caps = get_caps_db(db, geo)
    del caps['inflow']
    n_out = len(caps)

    # time constants (analytical)
    const['tau']['ana'] = np.array([tau_ana_dic[c] for c in caps])

    # factor between asymptotic and cyclic error (analytical)
    const['alpha']['ana'] = 1 / (np.exp(1 / const['tau']['ana']) - 1)

    const['tau']['num'] = {}
    const['alpha']['num'] = {}

    # tolerance for the linear slope in a log-plot
    tol = {'pressure': 1e-10,
           'flow': 1e-6}

    for f in c_res.keys():
        # # chose range of cardiac cycles to evaluate time constants (within given tolerance)
        i_min = -1
        i_good = np.sum(np.abs(np.diff(-np.diff(np.log(c_res[f]['cyclic']), axis=0), axis=0)) < tol[f], axis=1) == n_out
        for i, g in enumerate(i_good):
            if i_min == -1 and g:
                i_min = i
                continue
            if i_min > -1 and not g:
                i_max = i - 1
                break
        else:
            i_min = 0
            i_max = np.min([10, c_res[f]['cyclic'].shape[0] - 1])
        i_calc = np.arange(i_min, i_max)

        # time constants (numerical)
        const['tau']['num'][f] = 1 / np.mean(-np.diff(np.log(c_res[f]['cyclic'][i_calc]), axis=0), axis=0)

        # factor between asymptotic and cyclic error (numerical)
        const['alpha']['num'][f] = np.mean(c_res[f]['asymptotic'][i_calc + 1] / c_res[f]['cyclic'][i_calc], axis=0)

    return const


def plot_pressure(study, geo):
    # plot settings
    m = '3d_rerun'
    # m = '1d'
    # m = '0d'
    fields = ['pressure', 'flow']
    # fields = ['pressure']
    # comparisons = ['cycle', 'initial', 'cyclic', 'asymptotic']
    # comparisons = ['cycle', 'mean', 'asymptotic']
    # comparisons = ['cycle', 'mean', 'cyclic', 'asymptotic']
    comparisons = ['cycle', 'mean', 'cyclic', 'asymptotic']
    # comparisons = ['cycle_norm', 'mean', 'cyclic', 'asymptotic']

    # get database
    db = Database(study)

    # res_m, time = collect_results_db_1d_3d(db, geo)
    if m == '0d':
        res_m, time = collect_results_db_0d(db, geo)
    elif m == '3d_rerun':
        res_m, time = collect_results_db_3d_3d(db, geo)
    else:
        return

    if res_m is None:
        return

    if os.path.exists(db.get_bc_0D_path(geo, m)):
        res_0d = np.load(db.get_bc_0D_path(geo, m), allow_pickle=True).item()
    else:
        return

    print(geo)
    errors = collect_errors(res_m, res_0d, time, m)

    fig, ax = plt.subplots(len(fields), len(comparisons), figsize=(5 * len(comparisons), 5 * len(fields)), dpi=300)

    c_res = defaultdict(dict)
    i_conv = {}
    for j, f in enumerate(fields):
        # for i, p in zip([0, 1, 2, 2], comparisons):
        for i, p in enumerate(comparisons):
            if len(fields) == 1:
                pos = i
            else:
                pos = (j, i)

            c_res[f][p], i_conv[f] = make_err_plot(db, geo, ax, pos, m, f, p, res_m, errors, time)

    f_out = db.get_statistics_dir()
    fname = 'convergence_' + study + '_' + geo
    if len(fields) == 1:
        fname += '_' + f
    fpath = os.path.join(f_out, fname + '.png')
    # plt.subplots_adjust(right=0.8)
    fig.tight_layout()
    fig.savefig(fpath, bbox_inches='tight')
    plt.close(fig)

    # plot time constants
    const = get_time_constants(db, geo, c_res)
    for f in fields:
        for c, v in const.items():
            fig, ax = plt.subplots(figsize=(6, 4), dpi=300)
            ax.plot(const[c]['ana'], 'o')
            ax.plot(const[c]['num'][f], 'x')
            ax.grid('both')
            ax.legend(['analytical', 'numerical'])
            ax.set_xlabel('Outlet')
            ax.set_ylabel(c)
            os.makedirs(os.path.join(f_out, c), exist_ok=True)
            fpath = os.path.join(f_out, c, c + '_' + study + '_' + geo + '_' + f + '.png')
            fig.savefig(fpath, bbox_inches='tight')
            plt.close(fig)

    return {'tau': const['tau']['num'], 'i_conv': i_conv}


def plot_pressure_studies(geo):
    # plot settings
    m = '3d_rerun'
    fields = ['pressure', 'flow']
    # studies = ['ini_zero', 'ini_steady', 'ini_1d_quad']
    # studies = ['ini_zero', 'ini_steady', 'ini_irene', 'ini_1d_quad']
    studies = ['ini_zero', 'ini_irene', 'ini_1d_quad']
    # studies = ['ini_1d_quad', 'ini_asymp_pres_1d_velo', 'ini_1d_pres_asymp_velo']
    comparison = 'asymptotic'
    # comparison = 'mean'

    print(geo)
    fig, ax = plt.subplots(len(fields), len(studies), figsize=(6 * len(studies), 5 * len(fields)), dpi=300, sharey='row')

    for j, f in enumerate(fields):
        for i, p in enumerate(studies):
            # get database
            db = Database(p)

            if os.path.exists(db.get_bc_0D_path(geo, m)):
                res_0d = np.load(db.get_bc_0D_path(geo, m), allow_pickle=True).item()
            else:
                return

            res_m, time = collect_results_db_3d_3d(db, geo)
            errors = collect_errors(res_m, res_0d, time, m)

            if len(fields) == 1:
                pos = i
            else:
                pos = (j, i)

            make_err_plot(db, geo, ax, pos, m, f, comparison, res_m, errors, time, title_study=p)

    f_out = '/home/pfaller/work/paper/asymptotic'
    fname = 'comparison_' + geo
    if len(fields) == 1:
        fname += '_' + f
    fpath = os.path.join(f_out, fname + '.png')
    # plt.subplots_adjust(right=0.8)
    fig.tight_layout()
    fig.savefig(fpath, bbox_inches='tight')
    plt.close(fig)


def main(db, geo):
    for g in geo:
        # plot_pressure_studies(g)
        res = plot_pressure(db.study, g)
        if res is not None:
            db.add_convergence(g, res)


if __name__ == '__main__':
    descr = 'Make plots for 3D-1D-0D paper'
    d, g, _ = input_args(descr)
    # main(d, g)
    # plot_convergence(d)
    plot_correlation(d)
